"""
Convert the MTCNN weights from FaceNet-PyTorch's format into the structure used
by this library.
"""

from argparse import ArgumentParser

import torch
import numpy as np

import pickle

from faceie.mtcnn.model import (
    PNetWeights,
    RNetWeights,
    ONetWeights,
    Conv2DWeights,
    LinearWeights,
)


def main():
    parser = ArgumentParser(
        description="""
        Convert MTCNN weights from FaceNet-PyTorch format to that used by
        faceie.
        """
    )
    parser.add_argument(
        "--p-net",
        "-p",
        nargs=2,
        metavar=("input", "output"),
    )
    parser.add_argument(
        "--r-net",
        "-r",
        nargs=2,
        metavar=("input", "output"),
    )
    parser.add_argument(
        "--o-net",
        "-o",
        nargs=2,
        metavar=("input", "output"),
    )

    args = parser.parse_args()

    ################################################################################

    if args.p_net is not None:
        state_dict = torch.load(args.p_net[0])

        # The bounding box coordinates are given relative to the input top-left and
        # bottom-right corners respectively in the PyTorch implementation. I'd prefer
        # them both to be relative to the top-left corner coordinates
        bounding_boxes_biases = state_dict["conv4_2.bias"].numpy()  # (4)
        bounding_boxes_biases[2:] += 1

        weights = PNetWeights(
            conv1=Conv2DWeights(
                weights=state_dict["conv1.weight"].numpy(),
                biases=state_dict["conv1.bias"].numpy(),
            ),
            prelu1=state_dict["prelu1.weight"].numpy(),
            conv2=Conv2DWeights(
                weights=state_dict["conv2.weight"].numpy(),
                biases=state_dict["conv2.bias"].numpy(),
            ),
            prelu2=state_dict["prelu2.weight"].numpy(),
            conv3=Conv2DWeights(
                weights=state_dict["conv3.weight"].numpy(),
                biases=state_dict["conv3.bias"].numpy(),
            ),
            prelu3=state_dict["prelu3.weight"].numpy(),
            classifier=Conv2DWeights(
                weights=state_dict["conv4_1.weight"].numpy(),
                biases=state_dict["conv4_1.bias"].numpy(),
            ),
            bounding_boxes=Conv2DWeights(
                weights=state_dict["conv4_2.weight"].numpy(),
                biases=bounding_boxes_biases,
            ),
        )

        pickle.dump(weights, open(args.p_net[1], "wb"))

    ################################################################################

    if args.r_net is not None:
        state_dict = torch.load(args.r_net[0])

        # NB: The final dense layer in the PyTorch model assumes pixels are ordered
        # (width, height, channels) however, in our implementation by convention we use
        # (channels, height, width). As such we shuffle the matrix around now rather
        # than shuffling pixel data every time we use it...
        #
        # NB: Numpy linear transform weights stored transposed by PyTorch
        dense4_weight = state_dict["dense4.weight"].numpy()  # (128, 3*3*64)
        dense4_weight = dense4_weight.reshape(128, 3, 3, 64)  # (128, 3, 3, 64)
        dense4_weight = np.moveaxis(
            dense4_weight, (1, 2, 3), (3, 2, 1)
        )  # (128, 64, 3, 3)
        dense4_weight = dense4_weight.reshape(128, 64 * 3 * 3)  # (128, 64*3*3)
        dense4_weight = dense4_weight.T  # (64*3*3, 128)
        dense4_weight = np.ascontiguousarray(dense4_weight)

        # The bounding box coordinates are given relative to the input top-left and
        # bottom-right corners respectively in the PyTorch implementation. I'd prefer
        # them both to be relative to the top-left corner coordinates
        bounding_boxes_biases = state_dict["dense5_2.bias"].numpy()  # (4)
        bounding_boxes_biases[2:] += 1

        weights = RNetWeights(
            conv1=Conv2DWeights(
                weights=state_dict["conv1.weight"].numpy(),
                biases=state_dict["conv1.bias"].numpy(),
            ),
            prelu1=state_dict["prelu1.weight"].numpy(),
            conv2=Conv2DWeights(
                weights=state_dict["conv2.weight"].numpy(),
                biases=state_dict["conv2.bias"].numpy(),
            ),
            prelu2=state_dict["prelu2.weight"].numpy(),
            conv3=Conv2DWeights(
                weights=state_dict["conv3.weight"].numpy(),
                biases=state_dict["conv3.bias"].numpy(),
            ),
            prelu3=state_dict["prelu3.weight"].numpy(),
            linear=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                dense4_weight,
                biases=state_dict["dense4.bias"].numpy(),
            ),
            prelu4=state_dict["prelu4.weight"].numpy(),
            classifier=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                weights=state_dict["dense5_1.weight"].T.numpy(),
                biases=state_dict["dense5_1.bias"].numpy(),
            ),
            bounding_boxes=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                weights=state_dict["dense5_2.weight"].T.numpy(),
                biases=bounding_boxes_biases,
            ),
        )

        pickle.dump(weights, open(args.r_net[1], "wb"))

    ################################################################################

    if args.o_net is not None:
        state_dict = torch.load(args.o_net[0])

        # NB: The final dense layer in the PyTorch model assumes pixels are ordered
        # (width, height, channels) however, in our implementation by convention we use
        # (channels, height, width). As such we shuffle the matrix around now rather
        # than shuffling pixel data every time we use it...
        #
        # NB: Numpy linear transform weights stored transposed by PyTorch
        dense5_weight = state_dict["dense5.weight"].numpy()  # (256, 3*3*128)
        dense5_weight = dense5_weight.reshape(256, 3, 3, 128)  # (256, 3, 3, 128)
        dense5_weight = np.moveaxis(
            dense5_weight, (1, 2, 3), (3, 2, 1)
        )  # (256, 128, 3, 3)
        dense5_weight = dense5_weight.reshape(256, 128 * 3 * 3)  # (256, 128*3*3)
        dense5_weight = dense5_weight.T  # (128*3*3, 256)
        dense5_weight = np.ascontiguousarray(dense5_weight)

        # The bounding box coordinates are given relative to the input top-left and
        # bottom-right corners respectively in the PyTorch implementation. I'd prefer
        # them both to be relative to the top-left corner coordinates
        bounding_boxes_biases = state_dict["dense6_2.bias"].numpy()  # (4)
        bounding_boxes_biases[2:] += 1

        # NB: The dense6_3 transform orders its outputs as x1, x2, x3, x4, x5, y1, y2,
        # y3, y4, y5 which is inconsistent with the bounding box regressions which are
        # ordered x1, y1, x2, y2, etc. As a result we re-order the matrix here to
        # produce a better output ordering.
        #
        # NB: Numpy linear transform weights stored transposed by PyTorch
        dense6_3_weights = state_dict["dense6_3.weight"].T.numpy()  # (256, 10)
        dense6_3_weights = dense6_3_weights.reshape(256, 2, 5)  # (256, 2, 5)
        dense6_3_weights = np.moveaxis(dense6_3_weights, (1, 2), (2, 1))  # (256, 5, 2)
        dense6_3_weights = dense6_3_weights.reshape(256, 10)  # (256, 10) (reordered)

        dense6_3_biases = state_dict["dense6_3.bias"].numpy()  # (10)
        dense6_3_biases = dense6_3_biases.reshape(2, 5)  # (2, 5)
        dense6_3_biases = np.moveaxis(dense6_3_biases, 0, 1)  # (5, 2)
        dense6_3_biases = dense6_3_biases.reshape(10)  # (10) (reordered)

        weights = ONetWeights(
            conv1=Conv2DWeights(
                weights=state_dict["conv1.weight"].numpy(),
                biases=state_dict["conv1.bias"].numpy(),
            ),
            prelu1=state_dict["prelu1.weight"].numpy(),
            conv2=Conv2DWeights(
                weights=state_dict["conv2.weight"].numpy(),
                biases=state_dict["conv2.bias"].numpy(),
            ),
            prelu2=state_dict["prelu2.weight"].numpy(),
            conv3=Conv2DWeights(
                weights=state_dict["conv3.weight"].numpy(),
                biases=state_dict["conv3.bias"].numpy(),
            ),
            prelu3=state_dict["prelu3.weight"].numpy(),
            conv4=Conv2DWeights(
                weights=state_dict["conv4.weight"].numpy(),
                biases=state_dict["conv4.bias"].numpy(),
            ),
            prelu4=state_dict["prelu4.weight"].numpy(),
            linear=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                dense5_weight,
                biases=state_dict["dense5.bias"].numpy(),
            ),
            prelu5=state_dict["prelu5.weight"].numpy(),
            classifier=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                weights=state_dict["dense6_1.weight"].T.numpy(),
                biases=state_dict["dense6_1.bias"].numpy(),
            ),
            bounding_boxes=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                weights=state_dict["dense6_2.weight"].T.numpy(),
                biases=bounding_boxes_biases,
            ),
            landmarks=LinearWeights(
                # NB: Linear transform weights stored transposed by PyTorch
                weights=dense6_3_weights,
                biases=dense6_3_biases,
            ),
        )

        pickle.dump(weights, open(args.o_net[1], "wb"))



"""
An implementation of the Non-Maximum Suppression algorithm for removing
(approximate) duplicates from a set of overlapping rectangles.

.. autofunction:: non_maximum_suppression
"""

import numpy as np
from numpy.typing import NDArray


def non_maximum_suppression(
    probabilities: NDArray,
    bounding_boxes: NDArray,
    maximum_iou: float,
) -> NDArray:
    """
    Performs non-maximum suppression (NMS) on a collection of bounding boxes
    and associated probabilities. This returns the subset of bounding boxes
    which are both sufficiently non-overlapping.

    Parameters
    ==========
    probabilities : array (num_candidates)
        An array of probabilities of each candidate bounding box containing a
        face.
    bounding_boxes : array (num_candidates, 4)
        The corresponding bounding boxes which may contain a face, for the same
        shape of values as ``probabilities``. The last axis contains the
        bounding box corners as x1, y1, x2 and y2 respectively.
    maximum_iou : float
        The maximum overlapping area (as Intersection over Union (IoU)) two
        bounding boxes can have before the lower-probability box is removed.

    Returns
    =======
    array (num_results)
        An array of indices into the input ``probabilities`` and
        ``bounding_boxes`` arrays which indicate the bounding boxes selected by
        the non-maximum suppression algoirthm.
    """
    # We proceed by repeatedly selecting the most likely bounding box and then
    # removing any others which overlap too much with it.
    selected_indices = []
    candidate_indices = np.argsort(probabilities)
    while len(candidate_indices) > 0:
        # Find (and keep) most likley face
        best_index = candidate_indices[-1]
        selected_indices.append(best_index)
        candidate_indices = candidate_indices[:-1]

        # Now lets find (and remove) any faces which overlap with this face by
        # too much.

        # Bounding box of most likely ('Best') face
        bx1, by1, bx2, by2 = bounding_boxes[best_index]

        # Bounding boxes of all 'Other' candidate faces
        ox1s, oy1s, ox2s, oy2s = bounding_boxes[candidate_indices].T

        # Work out intersecting areas
        #
        # In 1D, consider two pairs of coordinates with an intersecting region:
        #
        #     ax1|----------|ax2
        #         bx1|--------------|bx2
        #         ix1|------|ix2
        #
        # The intersecting region (denoted by ix1 and ix2 above) is given by
        #
        #     ix1 = max(ax1, bx1)
        #     ix2 = min(ax2, bx2)
        #
        # There is an intersection whenever ix1 < ix2, otherwise the regions do
        # not intersect.
        #
        # So, lets carry this out en-masse on our bounding box vs all other
        # bounding boxes -- and for both x and y axes:
        intersection_x1s = np.maximum(bx1, ox1s)
        intersection_x2s = np.minimum(bx2, ox2s)
        intersection_xs = np.maximum(0, intersection_x2s - intersection_x1s)

        intersection_y1s = np.maximum(by1, oy1s)
        intersection_y2s = np.minimum(by2, oy2s)
        intersection_ys = np.maximum(0, intersection_y2s - intersection_y1s)

        intersection_areas = intersection_xs * intersection_ys

        # Now lets compute the total area of the union of the best bounding box
        # and all other bounding boxes.
        best_area = (bx2 - bx1) * (by2 - by1)
        other_areas = (ox2s - ox1s) * (oy2s - oy1s)
        union_areas = best_area + other_areas - intersection_areas

        # Finally, lets compute the intersection-over-union (IoU) and discard
        # any candidates with excessive overlap
        ious = intersection_areas / union_areas
        candidate_indices = candidate_indices[np.where(ious <= maximum_iou)]

    return np.array(selected_indices, dtype=int)



"""
A (Numpy-based) implementation of face-to-vector encoding based on the
FaceNet methodology but using an Inception-ResNet-v1 network architecture
based on "FaceNet-PyTorch" (https://github.com/timesler/facenet-pytorch).

Background
==========

The FaceNet methodology for producing a face-to-vector system was introduced in
"FaceNet: A Unified Embedding for Face Recognition and Clustering" by Schroff,
Kalenichenko and Philbin (2015). It describes a collection of networks,
including several based on the 'Inception-v1' convolutional network
architecture capable of producing state-of-the-art face vector embeddings
suitable for various tasks.

The Inception-ResNet-v1 architecture was introduced in "Inception-v4,
Inception-ResNet and the Impact of Residual Connections on Learning" by
Szegedy, Ioffe, Vanhoucke and Alemi (2016). These Inception architectures form
the forth generation of refinements boasting improvements in performance and
training speed.

This implementation is intended to be weight-compatible with Tim Esler's
FaceNet-PyTorch implementation (https://github.com/timesler/facenet-pytorch).
This, in turn, was built on weights ported from David Sandberg's FaceNet
implementation based on Tensorflow (https://github.com/davidsandberg/facenet).
The latter began life as an implementation of the FaceNet paper's Inception-v1
based "NN3" model, hence the 'FaceNet' name. Later on, this project switched
over to using the Inception-ResNet-v1 architecture. It is this network which
FaceNet-PyTorch eventually re-implemented and so to what this module
reimplements using Numpy. Inception indeed.

"""

from faceie.facenet.model import encode_faces, FaceNetWeights, load



"""
This submodule implements the three deep convolutional networks from the MTCNN
face detection system. These are:

.. autofunction:: p_net
.. autofunction:: r_net
.. autofunction:: o_net

Also included in this module are the various datastructures used to hold the
model weights for these functions:

.. autoclass:: PNetWeights
.. autoclass:: RNetWeights
.. autoclass:: ONetWeights
.. autoclass:: LinearWeights
.. autoclass:: Conv2DWeights

The various ``*NetWeights`` classes include a ``load()`` class method which
loads a copy of these weights from the ``data`` subdirectory of this module.
The weights are simply stored pickled on disk since they are small enough not
to warrant any special handling.
"""

from typing import NamedTuple, Iterator, cast
from numpy.typing import NDArray

from functools import cache

import pickle

from pathlib import Path

import numpy as np

from faceie.nn import (
    linear,
    conv2d,
    max_pool_2d,
    softmax,
    prelu,
)

from faceie.nn_arguments import (
    LinearWeights,
    Conv2DWeights,
)


MODEL_DATA_DIR = Path(__file__).parent / "data"
"""
Directory containing precomputed model weights files.
"""


class PNetWeights(NamedTuple):
    """
    Weghts for the :py:func:`p_net` 'proposal' network.
    """

    # First 3x3 convolution 3-channels in, 10-channels out
    conv1: Conv2DWeights
    prelu1: NDArray

    # Second 3x3 convolution 10-channels in, 16-channels out
    conv2: Conv2DWeights
    prelu2: NDArray

    # Third 3x3 convolution 16-channels in, 32-channels out
    conv3: Conv2DWeights
    prelu3: NDArray

    # Output 1x1 'convolutions' (really just simple matrix multiplies)
    classifier: Conv2DWeights  # 32-channels in, 2 channels out
    bounding_boxes: Conv2DWeights  # 32-channels in, 4 channels out

    @classmethod
    @cache
    def load(cls) -> "PNetWeights":
        filename = MODEL_DATA_DIR / "p_net_weights.pickle"
        return cast(PNetWeights, pickle.load(filename.open("rb")))


class RNetWeights(NamedTuple):
    """
    Weghts for the :py:func:`r_net` 'proposal' network.
    """

    # First 3x3 convolution 3-channels in, 28-channels out
    conv1: Conv2DWeights
    prelu1: NDArray

    # Second 3x3 convolution 28-channels in, 48-channels out
    conv2: Conv2DWeights
    prelu2: NDArray

    # Third 3x3 convolution 48-channels in, 64-channels out
    conv3: Conv2DWeights
    prelu3: NDArray

    # Fully-connected linear stage 64x3x3 in, 128 out.
    linear: LinearWeights
    prelu4: NDArray

    # Final output weight matrices
    classifier: LinearWeights
    bounding_boxes: LinearWeights

    @classmethod
    @cache
    def load(cls) -> "RNetWeights":
        filename = MODEL_DATA_DIR / "r_net_weights.pickle"
        return cast(RNetWeights, pickle.load(filename.open("rb")))


class ONetWeights(NamedTuple):
    """
    Weghts for the :py:func:`o_net` 'proposal' network.
    """

    # First 3x3 convolution 3-channels in, 32-channels out
    conv1: Conv2DWeights
    prelu1: NDArray

    # Second 3x3 convolution 32-channels in, 64-channels out
    conv2: Conv2DWeights
    prelu2: NDArray

    # Third 3x3 convolution 64-channels in, 64-channels out
    conv3: Conv2DWeights
    prelu3: NDArray

    # Fourth 2x2 convolution 64-channels in, 128-channels out
    conv4: Conv2DWeights
    prelu4: NDArray

    # Fully-connected linear stage 128x3x3 in, 256 out.
    linear: LinearWeights
    prelu5: NDArray

    # Final output weight matrices
    classifier: LinearWeights
    bounding_boxes: LinearWeights
    landmarks: LinearWeights

    @classmethod
    @cache
    def load(cls) -> "ONetWeights":
        filename = MODEL_DATA_DIR / "o_net_weights.pickle"
        return cast(ONetWeights, pickle.load(filename.open("rb")))


def p_net(img, weights: PNetWeights | None = None) -> tuple[NDArray, NDArray]:
    """
    The 'proposal' network which uses a convolutional nerual network to very
    quickly (but crudely) locate all potential faces in the input image.

    The convolution uses a 12x12 kernel operates with a 2D stride of (2, 2) on
    an un-padded input meaning that the output resolution is a little under
    half that of the input.  Specifically, output dimensions are (input
    dimension - 10) // 2. (NB: Whilst the output is strieded, all input pixels
    are used during processing).

    To be explicit, the mapping between convolution result coordinates and
    input image regions is like so:

    * Output (0, 0) uses inputs (0, 0) to (11, 11) inclusive,
    * Output (0, 1) uses inputs (0, 2) to (11, 13) inclusive,
    * Output (1, 0) uses inputs (2, 0) to (13, 11) inclusive.
    * ...

    Parameters
    ==========
    img : array (3, height, width)
        The input image for processing. Pixel values should be given in the
        range -1.0 to 1.0.
    weights : PNetWeights or None
        If omitted, a default set of weights will be loaded automatically.

    Returns
    =======
    probabilities : array (out_height, out_width)
        A probability between 0.0 and 1.0 of there being a face in the
        convolved region.
    bounding_boxes : array (out_height, out_width, 4)
        The bounding boxes of the faces (if any) for each value in
        probabilities.

        The four values in dimension 2 are x1, y1, x2, y2 respectively.  These
        values are also scaled down such that '0' is the top-left of the
        corresponding input area and '1' is the bottom-right.

        See also :py:func:`resolve_p_net_bounding_boxes`.
    """
    if weights is None:
        weights = PNetWeights.load()

    # First (3x3) convolution stage
    x = conv2d(img, *weights.conv1)  # (10, height-2, width-2)

    # NB: The Zhang et al. paper does not actually specify any particular
    # non-linearity and defer details to "Multi-view face detection using deep
    # convolutional neural networks", Farfade et al. (2015). This paper in turn
    # cites "Imagenet classification with deep convolutional neural networks.",
    # Krizhevsky et al. (2012) -- better known as the AlexNet paper -- who
    # use ReLU non-linearities. Nevertheless, in their published Matlab code,
    # PReLU is used and so we do too!
    x = prelu(x, weights.prelu1, axis=0)

    # NB: The Zhang et al. paper specifies 3x3 max pooling here but the PyTorch
    # implementation uses a 2x2 kernel instead. We do the same here to keep the
    # implementations compatible.
    x = max_pool_2d(
        x, kernel=2, stride=2, ceil_mode=True
    )  # (10, (height-2) // 2, (width-2) // 2)

    # Second (3x3) convolution stage
    x = conv2d(x, *weights.conv2)  # (16, ((height-2) // 2) - 2, ((width-2) // 2) - 2)
    # (16, ((height-6) // 2),     ((width-6) // 2))
    x = prelu(x, weights.prelu2, axis=0)

    # Final (3x3) convolution stage
    x = conv2d(x, *weights.conv3)  # (32, ((height-6)  // 2) - 2, ((width-6)  // 2) - 2)
    # (32, ((height-10) // 2),     ((width-10) // 2))
    # (32, out_height,             out_width)
    x = prelu(x, weights.prelu3, axis=0)

    # After the prior convolution and maximum pooling steps, the resulting
    # convolution filter has a size of 12x12 and a step size of 2x2 (due to the
    # max pooling) as illustrated in figure 2 in the Zhang et al. paper.

    # Classification (i.e. face probability)
    #
    # NB: This is actually just a simple matrix multiply from each 32-element
    # vector to a 2-element vector (two logits representing not-face and
    # is-face cases, respectively). This is implemented here via conv2d for
    # convenience using a 1x1 kernel rather than faffing around shuffling
    # indices about.
    classification = conv2d(x, *weights.classifier)  # (2, out_height, out_width)

    # Convert from pair of logits to simple "is face" probability
    probabilities = softmax(classification, axis=0)[1]  # (out_height, out_width)

    # Bounding box calculation
    bounding_boxes = conv2d(x, *weights.bounding_boxes)  # (4, out_height, out_width)
    bounding_boxes = np.moveaxis(bounding_boxes, 0, 2)  # (out_height, out_width, 4)

    return (probabilities, bounding_boxes)


def r_net(img, weights: RNetWeights | None = None) -> tuple[NDArray, NDArray]:
    """
    The 'refinement' network which uses a convolutional nerual network to
    refine the bounding box around a face detected by the :py:func:`p_net`.

    Parameters
    ==========
    img : array (num_batches, 3, 24, 24)
        A series of 24x24 pixel input images for processing. Pixel values
        should be given in the range -1.0 to 1.0.
    weights : RNetWeights or None
        If omitted, a default set of weights will be loaded automatically.

    Returns
    =======
    probabilities : array (num_batches)
        A probability between 0.0 and 1.0 of there being a face in each input
        image.
    bounding_boxes : array (num_batches, 4)
        A refined bounding box for the face within each image.

        The four values in dimension 1 are x1, y1, x2, y2 respectively.
        These values are also scaled down such that '0' is the top-left of the
        input and '1' is the bottom-right.
    """
    if weights is None:
        weights = RNetWeights.load()

    # Sanity check input dimension
    assert img.shape[-3:] == (3, 24, 24)

    # First (3x3) convolution stage
    x = conv2d(img, *weights.conv1)  # (num_batches, 28, 22, 22)
    x = prelu(x, weights.prelu1, axis=1)  # See PReLU/ReLU note in p_net
    x = max_pool_2d(x, kernel=3, stride=2, ceil_mode=True)  # (num_batches, 28, 11, 11)

    # Second (3x3) convolution stage
    x = conv2d(x, *weights.conv2)  # (num_batches, 48, 9, 9)
    x = prelu(x, weights.prelu2, axis=1)
    x = max_pool_2d(x, kernel=3, stride=2, ceil_mode=True)  # (num_batches, 48, 4, 4)

    # Third (2x2) convolution stage
    #
    # NB: A typographical error in Fig 2. of the Zhang et al. paper appears to
    # indicate an output channel count of 64128, however a space is simply
    # mising between the '64' and '128'.
    x = conv2d(x, *weights.conv3)  # (num_batches, 64, 3, 3)
    x = prelu(x, weights.prelu3, axis=1)

    # Final 'fully connected' stage reduces the input images to 128-dimensional
    # vectors
    #
    # NB: In the PyTorch implementation (that we're aiming for weight
    # compatibility with), their matrix assumes our pixel data is ordered as
    # (width, height, channels) and they shuffle this data around accordingly
    # during processing.  Instead, we use a reorganised weight matrix so that
    # we can put in the (channels, height, width) ordered data we have
    # directly.
    x = x.reshape(x.shape[0], 64 * 3 * 3)  # (num_batches, 64*3*3)
    x = linear(x, *weights.linear)  # (num_batches, 128)
    x = prelu(x, weights.prelu4, axis=1)

    # Classification (i.e. face probability)
    classification = linear(x, *weights.classifier)  # (num_batches, 2)
    probabilities = softmax(classification, axis=-1)[..., 1]  # (num_batches, )

    # Bounding boxes
    bounding_boxes = linear(x, *weights.bounding_boxes)  # (num_batches, 4)

    return (probabilities, bounding_boxes)


def o_net(img, weights: ONetWeights | None = None) -> tuple[NDArray, NDArray, NDArray]:
    """
    The 'output' network which uses a convolutional nerual network to finally
    define the bounding box, facial features and probabilities for a face
    detected by the :py:func:`r_net`.

    Parameters
    ==========
    img : array (num_batches, 3, 48, 48)
        A series of 48x48 pixel input images for processing. Pixel values
        should be given in the range -1.0 to 1.0.
    weights : ONetWeights or None
        If omitted, a default set of weights will be loaded automatically.

    Returns
    =======
    probabilities : array (num_batches)
        A probability between 0.0 and 1.0 of there being a face in each input
        image.
    bounding_boxes : array (num_batches, 4)
        A refined bounding box for the face within each image.

        The four values in dimension 1 are x1, y1, x2, y2 respectively.
        These values are also scaled down such that '0' is the top-left of the
        input and '1' is the bottom-right.
    landmarks : array (num_batches, 10)
        The facial feature landmarks for the face within the image.

        The ten values in dimension 1 are x1, y1, x2, y2 and so on where:

        * (x1, y1) is the left eye coordinate
        * (x2, y2) is the right eye coordinate
        * (x3, y3) is the nose coordinate
        * (x4, y4) is the left mouth coordinate
        * (x5, y5) is the right mouth coordinate

        All values are scaled such that '1' is the width or height of the input
        bounding box and are relative to the top-left corner of the input.
    """
    if weights is None:
        weights = ONetWeights.load()

    # Sanity check input dimension
    assert img.shape[-3:] == (3, 48, 48)

    # First (3x3) convolution stage
    x = conv2d(img, *weights.conv1)  # (num_batches, 32, 46, 46)
    x = prelu(x, weights.prelu1, axis=1)  # See PReLU/ReLU note in p_net
    x = max_pool_2d(x, kernel=3, stride=2, ceil_mode=True)  # (num_batches, 32, 23, 23)

    # Second (3x3) convolution stage
    x = conv2d(x, *weights.conv2)  # (num_batches, 64, 21, 21)
    x = prelu(x, weights.prelu2, axis=1)
    x = max_pool_2d(x, kernel=3, stride=2, ceil_mode=True)  # (num_batches, 64, 10, 10)

    # Third (3x3) convolution stage
    x = conv2d(x, *weights.conv3)  # (num_batches, 64, 8, 8)
    x = prelu(x, weights.prelu3, axis=1)
    x = max_pool_2d(x, kernel=2, stride=2, ceil_mode=True)  # (num_batches, 64, 4, 4)

    # Fourth (2x2) convolution stage
    x = conv2d(x, *weights.conv4)  # (num_batches, 128, 3, 3)
    x = prelu(x, weights.prelu4, axis=1)

    # Final 'fully connected' stage reduces the input images to 256-dimensional
    # vectors.
    #
    # NB: Matrix re-ordered to match image memory layout compared with PyTorch
    # implementation -- see comment in r_net's fully connected stage.
    x = x.reshape(x.shape[0], 128 * 3 * 3)  # (num_batches, 128*3*3)
    x = linear(x, *weights.linear)  # (num_batches, 256)
    x = prelu(x, weights.prelu5, axis=1)

    # Classification (i.e. face probability)
    classification = linear(x, *weights.classifier)  # (num_batches, 2)
    probabilities = softmax(classification, axis=-1)[..., 1]  # (num_batches, )

    # Bounding boxes
    bounding_boxes = linear(x, *weights.bounding_boxes)  # (num_batches, 4)

    # Facial feature landmarks
    #
    # NB: Matrix is re-ordered compared with PyTorch implementation to
    # interleave x and y coordinates of the landmarks rather than having them
    # as 5 x coordinates followed by 5 y coordinates.
    landmarks = linear(x, *weights.landmarks)  # (num_batches, 10)

    return (probabilities, bounding_boxes, landmarks)



"""
Containers for storing the arguments to various functions in
:py:mod:`faceie.nn`.
"""


from typing import NamedTuple

from numpy.typing import NDArray


class LinearWeights(NamedTuple):
    """
    Linear layer weights, as used by :py:func:`faceie.nn.linear`.
    """

    weights: NDArray
    biases: NDArray | None


class Conv2DWeights(NamedTuple):
    """
    Convolutional kernel and biases for use by :py:func:`faceie.nn.conv2d`.
    """

    weights: NDArray
    biases: NDArray | None


class BatchNormalisationWeights(NamedTuple):
    """
    The mean, variance, weights and biases learnt during batch normalisation,
    as used by :py:func:`faceie.nn.batch_normalisation_2d`
    """

    population_mean: NDArray
    population_variance: NDArray

    weights: NDArray
    biases: NDArray

    eps: float



"""
A (Numpy-based) implementation of face detection based on Multi-Task Cascaded
Convolutional Networks (MTCNN).

.. autofunction:: detect_faces

.. autoclass:: DetectedFaces


Background
==========

This technique was originally published by Zhang et al. in "Joint Face
Detection and Alignment using Multi-task Cascaded Convolutional Networks"
(2016). This implementation attempts to be weight-compatible with Facenet
PyTorch (https://github.com/timesler/facenet-pytorch) since we use their
weights. In all other respects, this implementation is independent of others
and does not reuse any code.

Whilst the (internal) convolutional networks do produce identical results to
Facenet-PyTorch's networks, the other parts of the implementation differ
slightly either due to different default hyperparameters or just small
differences in implementation.

As an aside, I found that other implementations were frequently unclear and
difficult to follow (in my humble opinion, as an outsider). As a result, I've
tried to make this implementation somewhat easier to follow -- or at least more
explicit.


Performance
===========

Performance-wise this (CPU-only) implementation is somewhat slower
than the (CPU-based) one in Facenet-PyTorch by a factor of around 2-3. This
largely due to certain Numpy operations (in particular ``np.amax`` used in the
max pooling layers) not automatically parallelising across CPU cores. Inserting
some fairly crude parallelism brought the speed difference down to 1-2 times
slower but since I'd rather keep things clean (it is already "fast enough" for
my uses) I've not pursued this line.
"""

from faceie.mtcnn.detect_faces import detect_faces, DetectedFaces



"""
An implementation of a simple image pyramid structure.

.. autoclass:: ImagePyramid
"""

from typing import Iterator

from PIL import Image

import math


class ImagePyramid:
    """
    Implements a simple image 'pyramid' containing multiple copies of an input
    image scaled down by a factor of two until a particular minimum size.
    """

    # The full set of images in the pyramid. The i-th image is a factor of 2^i
    # smaller than the input image.
    _images: list[Image.Image]

    def __init__(
        self, image: Image.Image, min_size: int, downscale_factor: float = 2.0
    ) -> None:
        """
        Parameters
        ==========
        image : PIL Image
            The full-sized original image to use as the 'base' of the pyramid.
        downscale_factor : double
            The down-scaling factor applied to each successive image in the
            pyramdid.
        min_size : int
            The minimum height or width required for the smallest image in the
            pyramid. If the input is smaller than this a :py:exc:`ValueError`
            will be thrown.
        """
        if image.size[0] < min_size or image.size[1] < min_size:
            raise ValueError("Image smaller than minimum size.")

        assert downscale_factor > 1.0
        self.downscale_factor = downscale_factor

        # Create pyramid
        self._images = [image]
        while True:
            w, h = image.size
            w = round(w / self.downscale_factor)
            h = round(h / self.downscale_factor)
            if w < min_size or h < min_size:
                break

            image = image.resize((w, h))
            self._images.append(image)

    def __len__(self) -> int:
        """Get the number of levels in the pyramid."""
        return len(self._images)

    def __getitem__(self, idx: int) -> Image.Image:
        """Return the image at the given pyramid level."""
        return self._images[idx]

    def __iter__(self) -> Iterator[Image.Image]:
        return iter(self._images)

    def scale_between(self, level_a: int, level_b: int) -> float:
        """
        Return the scaling factor from pixel coordinates at level_a to
        equivalent pixel coordinates at level_b.
        """
        return self.downscale_factor ** (level_a - level_b)

    def closest_level(self, downscale_factor: float) -> int:
        """
        Return the level index whose downscale factor from the original image
        is as small as possible whilst being at least downscale_factor.
        """
        return min(
            len(self) - 1,
            max(
                0,
                math.floor(
                    math.log(downscale_factor) / math.log(self.downscale_factor)
                ),
            ),
        )

    def extract(
        self,
        crop: tuple[float, float, float, float],
        size: tuple[int, int] | None = None,
    ) -> Image.Image:
        """
        Extract a region within the image at the specified resolution.

        Parameters
        ==========
        crop : (x1, y1, x2, y2)
            The input region to extract (in full-resolution pixel coordinates).
            The x2 and y2 coordinates are exclusive: that is, these will be the
            first column and row *not* included in the crop.
        size : (width, height) or None
            If given, specifies the size to scale the output to. Otherwise,
            the native size is assumed.
        """
        x1, y1, x2, y2 = crop
        iw = round(x2 - x1)
        ih = round(y2 - y1)

        if size is None:
            size = (iw, ih)
        ow, oh = size

        # Pick the smallest scale from the pyramid at least as high resolution
        # as the output size to make the rescaling operation as small (and as
        # cheap) as possible.
        source_level = self.closest_level(min(iw / ow, ih / oh))

        level_scale_factor = self.scale_between(0, source_level)
        x1 *= level_scale_factor
        y1 *= level_scale_factor
        x2 *= level_scale_factor
        y2 *= level_scale_factor

        extract = self[source_level].crop((round(x1), round(y1), round(x2), round(y2)))
        return extract.resize((ow, oh))



"""
Numpy based implementations of the main Inception-ResNet-v1/FaceNet model components.
"""

from typing import NamedTuple, Iterable, Callable

from PIL import Image

import numpy as np
from numpy.typing import NDArray

from pathlib import Path

from functools import lru_cache

import weightie

from faceie import __version__

from faceie.image_to_array import image_to_array

from faceie.nn import (
    linear,
    conv2d,
    relu,
    batch_normalisation_2d,
    max_pool_2d,
    PaddingType,
    l2_normalisation,
)

from faceie.nn_arguments import (
    LinearWeights,
    Conv2DWeights,
    BatchNormalisationWeights,
)


class ConvolutionalUnitWeights(NamedTuple):
    """
    Convolutional kernel and batch normalisation for use by the
    :py:func:`convolutional_unit` unit.
    """

    # Convolutional kernel weights
    #
    # (out_channels, in_channels, kernel_height, kernel_width)
    kernel: NDArray

    # Batch normalisation
    batch_normalisation: BatchNormalisationWeights


class StemWeights(NamedTuple):
    """
    Convolution weights used by the stem network.
    """

    conv_1: ConvolutionalUnitWeights
    conv_2: ConvolutionalUnitWeights
    conv_3: ConvolutionalUnitWeights
    conv_4: ConvolutionalUnitWeights
    conv_5: ConvolutionalUnitWeights
    conv_6: ConvolutionalUnitWeights


class InceptionResNetAWeights(NamedTuple):
    """
    Weights used for an Inception-ResNet-A module.
    """

    branch_1_conv: ConvolutionalUnitWeights

    branch_2_conv_1: ConvolutionalUnitWeights
    branch_2_conv_2: ConvolutionalUnitWeights

    branch_3_conv_1: ConvolutionalUnitWeights
    branch_3_conv_2: ConvolutionalUnitWeights
    branch_3_conv_3: ConvolutionalUnitWeights

    output_conv: Conv2DWeights

    # NB: This scaling factor is appled to the output of the above convolutions
    # prior to adding the input values. That is, smaller numbers reduce the
    # influence of the convolution results relative to the input in the final
    # output.
    residual_scale: float


class ReductionAWeights(NamedTuple):
    """
    Weights used for the Reduction-A module.
    """

    branch_1_conv: ConvolutionalUnitWeights

    branch_2_conv_1: ConvolutionalUnitWeights
    branch_2_conv_2: ConvolutionalUnitWeights
    branch_2_conv_3: ConvolutionalUnitWeights


class InceptionResNetBWeights(NamedTuple):
    """
    Weights used for an Inception-ResNet-B module.
    """

    branch_1_conv: ConvolutionalUnitWeights

    branch_2_conv_1: ConvolutionalUnitWeights
    branch_2_conv_2: ConvolutionalUnitWeights
    branch_2_conv_3: ConvolutionalUnitWeights

    output_conv: Conv2DWeights

    # NB: This scaling factor is appled to the output of the above convolutions
    # prior to adding the input values. That is, smaller numbers reduce the
    # influence of the convolution results relative to the input in the final
    # output.
    residual_scale: float


class ReductionBWeights(NamedTuple):
    """
    Weights used for the Reduction-B module.
    """

    branch_1_conv_1: ConvolutionalUnitWeights
    branch_1_conv_2: ConvolutionalUnitWeights

    branch_2_conv_1: ConvolutionalUnitWeights
    branch_2_conv_2: ConvolutionalUnitWeights

    branch_3_conv_1: ConvolutionalUnitWeights
    branch_3_conv_2: ConvolutionalUnitWeights
    branch_3_conv_3: ConvolutionalUnitWeights


class InceptionResNetCWeights(NamedTuple):
    """
    Weights used for an Inception-ResNet-C module.
    """

    branch_1_conv: ConvolutionalUnitWeights

    branch_2_conv_1: ConvolutionalUnitWeights
    branch_2_conv_2: ConvolutionalUnitWeights
    branch_2_conv_3: ConvolutionalUnitWeights

    output_conv: Conv2DWeights

    # NB: This scaling factor is appled to the output of the above convolutions
    # prior to adding the input values. That is, smaller numbers reduce the
    # influence of the convolution results relative to the input in the final
    # output.
    residual_scale: float


class FaceNetWeights(NamedTuple):
    """Weights for the complete face-to-embedding function."""

    stem: StemWeights

    inception_resnet_a: list[InceptionResNetAWeights]
    reduction_a: ReductionAWeights
    inception_resnet_b: list[InceptionResNetBWeights]
    reduction_b: ReductionBWeights
    inception_resnet_c: list[InceptionResNetCWeights]

    output_dimension_reduction: LinearWeights
    output_batch_normalisation: BatchNormalisationWeights


# NB: This cache means that when load is called repeatedly and automatically by
# the encode function below it will not actually require us to re-load the
# weights file.
#
# NB: since the returned data is mmaped anyway we don't need to worry too much
# about keeping the weights loaded beyond their useful lifetime as the OS can
# swap them out if they're getting in the way anyway.
@lru_cache(maxsize=1)
def load(
    source: Path | str = "20180402-114759-vggface2.weights",
    search_paths: list[Path] | None = None,
    update: bool = False,
    progress_callback: Callable[[list[str], str, int, int | None], None]
    | None = weightie.downloader.print_status,
    min_callback_interval: float = 0.5,
) -> FaceNetWeights:
    """
    Load a set of weights from a file (if a Path is given) or automatically
    download a named weights file from the GitHub release (if a string is
    given).

    Parameters
    ==========
    source : local weights file (Path) or GitHub asset filename (str)
        The weights to be downloaded.
    search_paths : [Path, ...] or None
        When the source is a GitHub asset filename, a list of locations to
        search for weights locally before resorting to downloading the file. If
        None is given, will search platform-specific data directories.

        When downloading weights, the first item on the search path (by default
        the user application data directory) will be used to store the
        downloaded weights.
    update : bool
        If True, force a check for new weights to download.
    progress_callback : f(list_of_files, file, bytes_downloaded, bytes_total) or None
        During file downloads this will be called every min_callback_interval
        seconds with the status of the download. By default this will print
        status to stderr. Disable by passing None.
    min_callback_interval : float
        See progress_callback.

    Returns
    =======
    Weights
        The model weights.

        Loaded weights are memory mmapped meaning that the data will not
        actually be read from disk until it is used and may be freely swapped
        out of RAM by the OS when needed if they're not being used.
    """
    if isinstance(source, str):
        source = weightie.download(
            repository="mossblaser/faceie",
            asset_filenames=[source],
            target_version=__version__,
            search_paths=search_paths,
            update=update,
            progress_callback=progress_callback,
            min_callback_interval=min_callback_interval,
        )[source]

    return weightie.load(source.open("rb"))


def convolutional_unit(
    x: NDArray,
    weights: ConvolutionalUnitWeights,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] | PaddingType = 0,
) -> NDArray:
    """
    Implements the base convolutional unit used throughout Inception-style
    networks::

                   +--------+      +------------+      +------+
        Input ---> | Conv2D | ---> | Batch Norm | ---> | ReLU | ---> Output
                   +--------+      +------------+      +------+

    Parameters
    ==========
    x : array (num_batches, in_channels, height, width)
    weights : ConvolutionalUnitWeights
    stride : int or (int, int)
        The stride of the convolution operation.
    padding : int or (int, int) or PaddingType
        The padding of the convolution operation.

    Returns
    =======
    array (batch, out_channels, height, width)
    """
    # NB: The convolution does not include a bias since the batch normalisation
    # includes its own bias.
    x = conv2d(x, weights.kernel, stride=stride, padding=padding)

    # TODO: The batch normalisation essentially boils down to a single
    # multiply-add. We could potentially fold this into the convolution weights
    # (and add some biases) by adjusting the weights. Something to consider!
    x = batch_normalisation_2d(x, *weights.batch_normalisation, channel_axis=1)

    x = relu(x)

    return x


def stem(img: NDArray, weights: StemWeights) -> NDArray:
    """
    Implements the 'stem' of the network as described by figure 14 of the
    Inception ResNet paper.

    The stem is intended to quickly drop the number of pixels (to about a 64th
    of the input) whist increasing the channel count (to 256).  This might be
    interpreted as performing low-level feature extraction from the input.

    Parameters
    ==========
    img : array (num_batches, 3, height, width)
        The input image(s).
    weights : StemWeights

    Returns
    =======
    array (num_batches, 256, out_height, out_width)
        A processed version of the input image with approximately a 64th of the
        number of pixels but 256 dimensions rather than three.

        The output dimensions are given by the following formulae:

        * out_height = ((img_height - 19) // 8)
        * out_width = ((img_width - 19) // 8)
    """
    x = img

    # 3x3 convolution, 3->32 channels, stride 2, no padding ('V')
    #
    # x_height = ((img_height - 3) // 2) + 1    # See conv2d output dim. formulae
    #             (img_height - 1) // 2
    # NB: x_width formulae are similar since all convolutions are square
    x = convolutional_unit(
        x, weights.conv_1, stride=2
    )  # (num_batches, 32, x_height, x_width)

    # 3x3 convolution, 32->32 channels, no padding ('V')
    #
    # x_height = ((img_height - 1) // 2) - 3 + 1
    #            ((img_height - 1) // 2) - 2
    #             (img_height - 5) // 2
    x = convolutional_unit(x, weights.conv_2)  # (num_batches, 32, x_height, x_width)

    # 3x3 convolution, 32->64 channels, zero padded to input size
    #
    # x_height = (img_height - 5) // 2    # Unchanged
    #
    # (num_batches, 64, x_height, x_width)
    x = convolutional_unit(x, weights.conv_3, padding=PaddingType.same)

    # 3x3 max pooling, stride 2, no padding ('V')
    #
    # x_height = ((((img_height - 5) // 2) - 3) // 2) + 1
    #             (((img_height - 5) // 2) - 1) // 2
    #             ( (img_height - 7) // 2)      // 2
    #               (img_height - 7) // 4
    x = max_pool_2d(x, kernel=3, stride=2)  # (num_batches, 64, x_height, x_width)

    # 1x1 convolution, 64->80 channels
    #
    # x_height = (img_height - 7) // 4
    #
    # (num_batches, 80, x_height, x_width)
    x = convolutional_unit(x, weights.conv_4, padding=PaddingType.same)

    # 3x3 convolution, 80->192 channels, no padding ('V')
    #
    # x_height =  (img_height - 7) // 4
    #            ((img_height - 7) // 4) - 3 + 1
    #            ((img_height - 7) // 4) - 2
    #             (img_height - 15) // 4
    x = convolutional_unit(x, weights.conv_5)  # (num_batches, 192, x_height, x_width)

    # 3x3 convolution, 192->256 channels, stride 2, no padding ('V')
    #
    # x_height =    (img_height - 15) // 4
    #            ((((img_height - 15) // 4) - 3) // 2) + 1
    #            ((((img_height - 15) // 4) - 1) // 2)
    #            ( ((img_height - 19) // 4)      // 2)
    #               (img_height - 19) // 8
    x = convolutional_unit(
        x, weights.conv_6, stride=2
    )  # (num_batches, 256, x_height, x_width)

    return x


def inception_resnet_a(img: NDArray, weights: InceptionResNetAWeights) -> NDArray:
    """
    Implements the Inception-ResNet-A module as described by figure 10 of the
    Inception ResNet paper.

    Parameters
    ==========
    img : array (num_batches, 256, height, width)
        The input image(s).
    weights : InceptionResNetAWeights

    Returns
    =======
    array (num_batches, 256, height, width)
        A processed version of the input with the same size and number of
        channels.
    """
    # Branch 1 (1x1 convolution)
    branch_1 = img
    # 1x1 convolution, 256->32 channels
    branch_1 = convolutional_unit(
        branch_1, weights.branch_1_conv, padding=PaddingType.same
    )

    # Branch 2 (3x3 convolution)
    branch_2 = img
    # 1x1 convolution, 256->32 channels
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_1, padding=PaddingType.same
    )
    # 3x3 convolution, 32->32 channels, zero padded to input size
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_2, padding=PaddingType.same
    )

    # Branch 3 (factorised 5x5 convolution)
    branch_3 = img
    # 1x1 convolution, 256->32 channels
    branch_3 = convolutional_unit(
        branch_3, weights.branch_3_conv_1, padding=PaddingType.same
    )
    # 3x3 convolution, 32->32 channels, zero padded to input size
    branch_3 = convolutional_unit(
        branch_3, weights.branch_3_conv_2, padding=PaddingType.same
    )
    # 3x3 convolution, 32->32 channels, zero padded to input size
    branch_3 = convolutional_unit(
        branch_3, weights.branch_3_conv_3, padding=PaddingType.same
    )

    # Combine branches
    output = np.concatenate(
        (branch_1, branch_2, branch_3), axis=1
    )  # (num_batches, 96, h, w)

    # 1x1 convolution, 96->256 channels
    output = conv2d(output, *weights.output_conv, padding=PaddingType.same)

    # Residual connection
    output *= weights.residual_scale
    output += img

    # Final ReLU
    output = relu(output)

    return output


def reduction_a(x: NDArray, weights: ReductionAWeights) -> NDArray:
    """
    Implements the Reduction-A module as described by figure 7 and table 1 of
    the Inception ResNet paper.

    This reduces the spatial resolution of the input whilst increasing the
    number of channels.

    Parameters
    ==========
    x : array (num_batches, 256, height, width)
        The input (as output by a :py:func:`inception_resnet_a` module).
    weights : ReductionAWeights

    Returns
    =======
    array (num_batches, 896, out_height, out_width)
        A processed version of the input with approximately a quarter of the
        number of pixels and 3.5 times the number of channels.

        The output dimensions are, more precisely, given by the formulae:

        * out_height = (height - 1) // 2
        * out_width = (width - 1) // 2
    """

    # NB: In this implementation, the branches are re-ordered with respect to
    # figure 7 to ensure weight compatibility with FaceNet-PyTorch. In the
    # paper, the max_poolling layer comes first, in FaceNet-PyTorch (and here)
    # it comes last.

    # Branch 1: 3x3 Convolution, 256 -> 384 channels, stride 2
    #
    # (num_batches, 384, out_height, out_width)
    branch_1 = convolutional_unit(x, weights.branch_1_conv, stride=2)

    # Branch 2: (factorised) 5x5 Convolution, 256 -> 256, stride 2
    #
    # 1x1 convolution, 256->192 channels
    # (num_batches, 192, height, width)
    branch_2 = convolutional_unit(x, weights.branch_2_conv_1, padding=PaddingType.same)
    # 3x3 convolution, 192->192 channels, zero padded to input size
    # (num_batches, 192, height, width)
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_2, padding=PaddingType.same
    )
    # 3x3 convolution, 192->256 channels, stride 2
    # (num_batches, 256, out_height, out_width)
    branch_2 = convolutional_unit(branch_2, weights.branch_2_conv_3, stride=2)

    # Branch 3: Simple max pooling with stride 2
    branch_3 = max_pool_2d(
        x, kernel=3, stride=2
    )  # (num_batches, 256, out_height, out_width)

    # (num_batches, 896, out_height, out_width)
    output = np.concatenate((branch_1, branch_2, branch_3), axis=1)

    return output


def inception_resnet_b(x: NDArray, weights: InceptionResNetBWeights) -> NDArray:
    """
    Implements the Inception-ResNet-B module as described by figure 11 of the
    Inception ResNet paper.

    Parameters
    ==========
    x : array (num_batches, 896, height, width)
        The input (as output by a :py:func:`reduction_a` module).
    weights : InceptionResNetBWeights

    Returns
    =======
    array (num_batches, 896, height, width)
        A processed version of the input with the same size and number of
        channels.
    """
    # Branch 1 (1x1 convolution)
    branch_1 = x
    # 1x1 convolution, 896->128 channels
    branch_1 = convolutional_unit(
        branch_1, weights.branch_1_conv, padding=PaddingType.same
    )

    # Branch 2 (factorised 7x7 convolution)
    branch_2 = x
    # 1x1 convolution, 896->128 channels
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_1, padding=PaddingType.same
    )
    # 1x7 convolution, 128->128 channels, zero padded to input size
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_2, padding=PaddingType.same
    )
    # 7x1 convolution, 128->128 channels, zero padded to input size
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_3, padding=PaddingType.same
    )

    # Combine branches
    output = np.concatenate((branch_1, branch_2), axis=1)  # (num_batches, 256, h, w)

    # 1x1 convolution, 256->896 channels
    output = conv2d(output, *weights.output_conv, padding=PaddingType.same)

    # Residual connection
    output *= weights.residual_scale
    output += x

    # Final ReLU
    output = relu(output)

    return output


def reduction_b(x: NDArray, weights: ReductionBWeights) -> NDArray:
    """
    Implements the Reduction-B module as described by figure 12 of the
    Inception ResNet paper.

    This further reduces the spatial resolution and increases the channel count
    of the input.

    Parameters
    ==========
    x : array (num_batches, 896, height, width)
        The input (as output by a :py:func:`inception_resnet_b` module).
    weights : ReductionBWeights

    Returns
    =======
    array (num_batches, 1792, out_height, out_width)
        A processed version of the input with approximately a quarter of the
        number of pixels and double the number of channels.

        The output dimensions are, more precisely, given by the formulae:

        * out_height = (height - 1) // 2
        * out_width = (width - 1) // 2
    """

    # NB: In this implementation, the branches are re-ordered with respect to
    # figure 12 to ensure weight compatibility with FaceNet-PyTorch. In the
    # paper, the max_poolling layer comes first, in FaceNet-PyTorch (and here)
    # it comes last.

    # Branch 1: 3x3 Convolution, 896 -> 384 channels, stride 2
    #
    # 1x1 convolution, 896->256 channels
    # (num_batches, 256, height, width)
    branch_1 = convolutional_unit(x, weights.branch_1_conv_1, padding=PaddingType.same)
    # 3x3 convolution, 256->384 channels, stride 2
    # (num_batches, 384, out_height, out_width)
    branch_1 = convolutional_unit(branch_1, weights.branch_1_conv_2, stride=2)

    # Branch 2: 3x3 Convolution, 896 -> 256 channels, stride 2
    #
    # 1x1 convolution, 896->256 channels
    # (num_batches, 256, height, width)
    branch_2 = convolutional_unit(x, weights.branch_2_conv_1, padding=PaddingType.same)
    # 3x3 convolution, 256->256 channels, stride 2
    # (num_batches, 256, out_height, out_width)
    branch_2 = convolutional_unit(branch_2, weights.branch_2_conv_2, stride=2)

    # Branch 3: (factorised) 5x5 Convolution, 256 -> 256, stride 2
    #
    # 1x1 convolution, 896->256 channels
    # (num_batches, 256, height, width)
    branch_3 = convolutional_unit(x, weights.branch_3_conv_1, padding=PaddingType.same)
    # 3x3 convolution, 256->256 channels, zero padded to input size
    # (num_batches, 256, height, width)
    branch_3 = convolutional_unit(
        branch_3, weights.branch_3_conv_2, padding=PaddingType.same
    )
    # 3x3 convolution, 256->256 channels, stride 2
    # (num_batches, 256, out_height, out_width)
    branch_3 = convolutional_unit(branch_3, weights.branch_3_conv_3, stride=2)

    # Branch 4: Simple max pooling with stride 2
    branch_4 = max_pool_2d(
        x, kernel=3, stride=2
    )  # (num_batches, 896, out_height, out_width)

    # (num_batches, 1792, out_height, out_width)
    output = np.concatenate((branch_1, branch_2, branch_3, branch_4), axis=1)

    return output


def inception_resnet_c(
    x: NDArray, weights: InceptionResNetCWeights, skip_relu: bool = False
) -> NDArray:
    """
    Implements the Inception-ResNet-C module as described by figure 13 of the
    Inception ResNet paper.

    Parameters
    ==========
    x : array (num_batches, 1792, height, width)
        The input (as output by a :py:func:`reduction_b` module).
    weights : InceptionResNetCWeights
    skip_relu : bool
        If True, the final ReLU operation prior to the output is omitted.

    Returns
    =======
    array (num_batches, 1792, height, width)
        A processed version of the input with the same size and number of
        channels.
    """
    # Branch 1 (1x1 convolution)
    branch_1 = x
    # 1x1 convolution, 1792->192 channels
    branch_1 = convolutional_unit(
        branch_1, weights.branch_1_conv, padding=PaddingType.same
    )

    # Branch 2 (factorised 3x3 convolution)
    branch_2 = x
    # 1x1 convolution, 1792->192 channels
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_1, padding=PaddingType.same
    )
    # 1x3 convolution, 192->192 channels, zero padded to input size
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_2, padding=PaddingType.same
    )
    # 3x1 convolution, 192->192 channels, zero padded to input size
    branch_2 = convolutional_unit(
        branch_2, weights.branch_2_conv_3, padding=PaddingType.same
    )

    # Combine branches
    output = np.concatenate((branch_1, branch_2), axis=1)  # (num_batches, 384, h, w)

    # 1x1 convolution, 384->1792 channels
    output = conv2d(output, *weights.output_conv, padding=PaddingType.same)

    # Residual connection
    output *= weights.residual_scale
    output += x

    # Final ReLU
    if not skip_relu:
        output = relu(output)

    return output


def encode_faces(
    image: Iterable[Image.Image] | Image.Image,
    weights: FaceNetWeights | None = None
) -> NDArray:
    """
    Produce the embeddings for a face, or series of faces.

    This implementation uses an adaption of the Inception-ResNet-v1
    architecture whose output stages are adapted to match the FaceNet
    architecture.

    Parameters
    ==========
    image : Image, [Image, ...]
        The face image or images.

        According to the Inception-ResNet-v1 paper, images should be 299x299
        pixels however FaceNet-PyTorch (which this implementation uses weights
        from) was trained on 160x160 images. As such, better results may be
        expected with the 160x160 images.
    weights : FaceNetWeights or None
        If None, a default set of weights will be downloaded using
        :py:func:`load`.

    Returns
    =======
    (512, ) or (num_batches, 512)
        A 512 dimensional embedding of the face or faces provided (depending on
        the shape of the input image.
    """
    if weights is None:
        weights = load()
    
    # Convert input to images
    if isinstance(image, Image.Image):
        x = image_to_array(image)
    else:
        x = np.stack(list(map(image_to_array, image)))

    # Force to (num_batches, 3, height, width)
    batched = x.ndim == 4
    if not batched:
        x = x.reshape(1, *x.shape)

    x = stem(x, weights.stem)

    for wa in weights.inception_resnet_a:
        x = inception_resnet_a(x, wa)

    x = reduction_a(x, weights.reduction_a)

    for wb in weights.inception_resnet_b:
        x = inception_resnet_b(x, wb)

    x = reduction_b(x, weights.reduction_b)

    # This implementation diverges from the Inception-ResNet-v1 paper at this
    # point in having an extra Inception-ResNet-C module without the ReLU stage
    # appended.
    for wc in weights.inception_resnet_c[:-1]:
        x = inception_resnet_c(x, wc)
    x = inception_resnet_c(x, weights.inception_resnet_c[-1], skip_relu=True)

    # "Adaptive average pooling" (or just taking the mean of all pixel values
    # in less fancy terms...)
    x = np.mean(x, axis=(2, 3))

    x = linear(x, *weights.output_dimension_reduction)

    x = batch_normalisation_2d(x, *weights.output_batch_normalisation, channel_axis=1)

    x = l2_normalisation(x, axis=1)

    if not batched:
        x = x[0]

    return x



"""
This submodule implements the overall face detection process
(:py:func:`detect_faces`).
"""


from typing import NamedTuple

from itertools import chain

import numpy as np
from numpy.typing import NDArray

from PIL import Image

from faceie.image_to_array import image_to_array
from faceie.mtcnn.pyramid import ImagePyramid
from faceie.mtcnn.model import p_net, r_net, o_net
from faceie.mtcnn.non_maximum_suppression import non_maximum_suppression


def resolve_p_net_bounding_boxes(bounding_boxes: NDArray) -> NDArray:
    """
    Given the bounding box output of :py:func:`p_net`, resolve all bounding box
    coordinates to absolute pixel coordinates in the input image.

    Modifies the provided array in-place, but also returns it for convenience.
    """
    # The convolution kernel is 12x12 and so coordinates run from 0 to 11
    # inclusive -- here we scale-up from the nominal range of 0-1.
    bounding_boxes *= 11.0

    # Since the p_net convolution has a stride of 2x2, the output grid is half
    # the size so we multiply by two here to compensate.
    ys, xs = np.indices(bounding_boxes.shape[:2]) * 2
    bounding_boxes[..., 0::2] += np.expand_dims(xs, -1)
    bounding_boxes[..., 1::2] += np.expand_dims(ys, -1)

    return bounding_boxes


def bounding_boxes_to_centers(bounding_boxes: NDArray) -> NDArray:
    """
    Given an array of bounding boxes, return their center coordinates.

    Parameters
    ==========
    bounding_boxes : array (..., 4)

    Returns
    =======
    array (..., 2)
    """
    return np.sum(bounding_boxes.reshape(-1, 2, 2), axis=-2) / 2.0


def bounding_boxes_to_longest_sides(bounding_boxes: NDArray) -> NDArray:
    """
    Given an array of bounding boxes, return their longest sides.

    Parameters
    ==========
    bounding_boxes : array (..., 4)

    Returns
    =======
    array (...)
    """
    lengths = np.diff(bounding_boxes.reshape(-1, 2, 2), axis=-2).reshape(-1, 2)
    return np.max(lengths, axis=-1)


def centers_to_square_bounding_boxes(centres: NDArray, sizes: NDArray) -> NDArray:
    """
    Given an array of suqare bounding box centers and their sizes, returns an
    array of bounding boxes.

    Parameters
    ==========
    centers: array (..., 2)
    sizes: array (...)

    Returns
    =======
    array (..., 4)
    """
    half_sizes = np.expand_dims(sizes / 2.0, -1)
    return np.concatenate(
        [
            centres - half_sizes,
            centres + half_sizes,
        ],
        axis=-1,
    )


def make_square(bounding_boxes: NDArray) -> NDArray:
    """
    Given an array of bounding boxes, enlarge these as necessary, centered on
    the existing centre point, to make them square.

    Parameters
    ==========
    bounding_boxes: array (..., 4)
        An array which contains a 4-long dimension enumerating the x1, y1, x2
        and y2 coordinates of a bounding box.

    Returns
    =======
    array (..., 4)
        A new bounding box array where all bounding boxes have x2-x1 == y2-y1.
    """
    centers = bounding_boxes_to_centers(bounding_boxes)
    longest_sides = bounding_boxes_to_longest_sides(bounding_boxes)
    return centers_to_square_bounding_boxes(centers, longest_sides)


def resolve_coordinates(
    input_bounding_boxes: NDArray, coordinate_pairs: NDArray
) -> NDArray:
    """
    Resolve coordinates (e.g. bounding box or landmark coordinates produced by
    the :py:func:`r_net` or :py:func:`o_net` functions) from 0-1 ranges to
    actual pixel coordinates.

    Parameters
    ==========
    input_bounding_boxes : array (num_batches, 4)
        For each entry in coordinate_pairs, the corresponding input pixel
        bounding box.
    coordinate_pairs : array (num_batches, even-number)
        An array of interleaved values like x1, y1, x2, y2, ... which will be
        scaled from nominal 0-1 ranges to actual pixel values.

        This array will be modified in-place.

    Returns
    =======
    Returns the ``coordinate_pairs`` array again.
    """
    widths = input_bounding_boxes[:, 2] - input_bounding_boxes[:, 0]
    heights = input_bounding_boxes[:, 3] - input_bounding_boxes[:, 1]

    # Scale offsets to input size
    coordinate_pairs[:, 0::2] *= np.expand_dims(widths, -1)
    coordinate_pairs[:, 1::2] *= np.expand_dims(heights, -1)

    # Translate into position
    coordinate_pairs[:, 0::2] += np.expand_dims(input_bounding_boxes[:, 0], -1)
    coordinate_pairs[:, 1::2] += np.expand_dims(input_bounding_boxes[:, 1], -1)

    return coordinate_pairs


def get_proposals(
    image: NDArray,
    probability_threshold: float = 0.7,
) -> tuple[NDArray, NDArray]:
    """
    Given an image, run the :py:func:`p_net` proposal stage against an image,
    returning the probabilities and bounding boxes of all potential faces
    detected.

    Parameters
    ==========
    image : array (3, height, width)
        The image to process as values between -1 and +1. Must be at least
        12x12 pixels.
    probability_threshold : float
        The minimum probability of a face to accept as a proposal.

    Return
    ======
    array (num_proposals)
        The probabilities assigned to each found face.
    array (num_proposals, 4)
        The bounding box for each found face. Dimension 1 gives (x1, y1, x2,
        y2) in image pixel coordinates. The x2 and y2 coordinates are
        inclusive.
    """
    probs, bboxes = p_net(image)

    # Flatten
    shape = probs.shape
    probs = probs.reshape(np.product(shape))
    bboxes = bboxes.reshape(np.product(shape), 4)

    # Select only candidates with sufficiently high probability
    above_threshold = probs > probability_threshold
    probs = probs[above_threshold]
    bboxes = bboxes[above_threshold]

    # Resolve bounding boxes to pixel coordinates
    bboxes *= 11
    ys, xs = np.unravel_index(np.flatnonzero(above_threshold), shape)
    bboxes[:, 0::2] += np.expand_dims(xs, 1) * 2
    bboxes[:, 1::2] += np.expand_dims(ys, 1) * 2

    return (probs, bboxes)


def refine_proposals(
    pyramid: ImagePyramid,
    bounding_boxes: NDArray,
    probability_threshold: float = 0.7,
) -> tuple[NDArray, NDArray]:
    """
    Given a series of potential faces, return a more accurate probability and
    bounding box using the :py:func:`r_net`.

    Parameters
    ==========
    pyramid : ImagePyramid
        The image in which the faces reside.
    bounding_boxes : array (num_proposals, 4)
        The bounding boxes defining the locations of faces in the input image.
    probability_threshold : float
        The minimum (refined) probability of a face to return.

    Return
    ======
    probabilities : array (num_accepted_proposals)
        The (refined) probabilities of the input faces which surpass the given
        probability_threshold.
    bounding_boxes : array (num_accepted_proposals, 4)
        For each accepted face, a new refined bounding box.  Dimension 1 gives
        (x1, y1, x2, y2) in image pixel coordinates. The x2 and y2 coordinates
        are inclusive.
    """
    crops = make_square(bounding_boxes)

    # Crop candidate faces from inputs
    faces = np.empty((crops.shape[0], 3, 24, 24), dtype=np.float32)
    for i, (x1, y1, x2, y2) in enumerate(crops):
        faces[i] = image_to_array(pyramid.extract((x1, y1, x2 + 1, y2 + 1), (24, 24)))

    probs, bboxes = r_net(faces)

    # Convert to pixel coordinates
    bboxes = resolve_coordinates(crops, bboxes)

    # Select high-probability candidates
    selection = probs >= probability_threshold
    probs = probs[selection]
    bboxes = bboxes[selection]

    return probs, bboxes


def output_proposals(
    pyramid: ImagePyramid,
    bounding_boxes: NDArray,
    probability_threshold: float = 0.7,
) -> tuple[NDArray, NDArray, NDArray]:
    """
    Given a series of potential faces, return a final (output) probability,
    bounding box and (optional) set of facial landmarks using the
    :py:func:`o_net`.

    Parameters
    ==========
    pyramid : ImagePyramid
        The image in which the faces reside.
    bounding_boxes : array (num_proposals, 4)
        The bounding boxes defining the locations of faces in the input image.
    probability_threshold : float
        The minimum (refined) probability of a face to return.

    Return
    ======
    probabilities : array (num_accepted_proposals)
        The (output) probabilities of the input faces which surpass the given
        probability_threshold.
    bounding_boxes : array (num_accepted_proposals, 4)
        For each accepted face, a new refined bounding box.  Dimension 1 gives
        (x1, y1, x2, y2) in image pixel coordinates. The x2 and y2 coordinates
        are inclusive.
    landmarks : array (num_accepted_proposals, 10)
        For each accepted face, the facial landmark coordinates. Dimension 1
        gives (x1, y1, x2, y2, ...) giving the coordinates of the left eye,
        right eye, nose, left mouth and right mouth.
    """
    crops = make_square(bounding_boxes)

    # Crop candidate faces from inputs
    faces = np.empty((crops.shape[0], 3, 48, 48), dtype=np.float32)
    for i, (x1, y1, x2, y2) in enumerate(crops):
        faces[i] = image_to_array(pyramid.extract((x1, y1, x2 + 1, y2 + 1), (48, 48)))

    probs, bboxes, landmarks = o_net(faces)

    # Convert to pixel coordinates
    bboxes = resolve_coordinates(crops, bboxes)
    landmarks = resolve_coordinates(crops, landmarks)

    # Select high-probability candidates
    selection = probs >= probability_threshold
    probs = probs[selection]
    bboxes = bboxes[selection]
    landmarks = landmarks[selection]

    return probs, bboxes, landmarks


def detect_faces_single_orientation(
    pyramid: ImagePyramid,
    proposal_skip_downscale_factor: float = 4.0,
    proposal_probability_threshold: float = 0.7,
    refine_probability_threshold: float = 0.7,
    output_probability_threshold: float = 0.7,
    proposal_maximum_iou: float = 0.5,
    refine_maximum_iou: float = 0.5,
    output_maximum_iou: float = 0.5,
) -> tuple[NDArray, NDArray, NDArray]:
    """
    Detect faces within an image using the MTCNN algorithm by Zhang et al.

    Parameters
    ==========
    pyramid : ImagePyramid
        An Image Pyramid with the image to detect faces within.
    proposal_skip_downscale_factor : float
        Skips running the proposal filtering process on images below this
        scaling factor. The choice of '4' as a default prevents the use of
        up-scaled faces in the refinement and output stages (whose input sizes
        are twice and four times that of the proposal stage).
    proposal_probability_threshold : float
    refine_probability_threshold : float
    output_probability_threshold : float
        The is-face probability threshold for a face to be accepted during each
        of the three filtering stages.
    proposal_maximum_iou : float
    refine_maximum_iou : float
    output_maximum_iou : float
        The threshold of the intersection-over-union above which two rectangles
        are duplicates during each of the three processing stages.

    Returns
    =======
    probabilities : array (num_accepted_proposals)
        The (output) probabilities of the input faces which surpass the given
        probability_threshold.
    bounding_boxes : array (num_accepted_proposals, 4)
        For each accepted face, a new refined bounding box.  Dimension 1 gives
        (x1, y1, x2, y2) in image pixel coordinates. The x2 and y2 coordinates
        are inclusive.
    landmarks : array (num_accepted_proposals, 10)
        For each accepted face, the facial landmark coordinates. Dimension 1
        gives (x1, y1, x2, y2, ...) giving the coordinates of the left eye,
        right eye, nose, left mouth and right mouth.
    """
    # First 'proposal' stage. Run convolution at multiple scales
    # ----------------------------------------------------------
    all_level_probs = []
    all_level_bboxes = []

    first_level = pyramid.closest_level(proposal_skip_downscale_factor)
    for level, image in enumerate(pyramid):
        if level < first_level or image.size[0] < 12 or image.size[1] < 12:
            continue

        probs, bboxes = get_proposals(
            image_to_array(image),
            proposal_probability_threshold,
        )

        # Remove overlaps within level using NMS
        selection = non_maximum_suppression(probs, bboxes, proposal_maximum_iou)

        scale_to_native = pyramid.scale_between(level, 0)
        all_level_probs.append(probs[selection])
        all_level_bboxes.append(bboxes[selection] * scale_to_native)

    probs = np.concatenate(all_level_probs)
    bboxes = np.concatenate(all_level_bboxes)

    # Remove overlaps between propsals at different scales using NMS on
    # aggregated mappings
    selection = non_maximum_suppression(probs, bboxes, proposal_maximum_iou)
    probs = probs[selection]
    bboxes = bboxes[selection]

    # Second 'refinement' stage. Run network against each image in turn.
    # ------------------------------------------------------------------
    probs, bboxes = refine_proposals(pyramid, bboxes, refine_probability_threshold)

    # Remove overlapping candidates with NMS
    selection = non_maximum_suppression(probs, bboxes, refine_maximum_iou)
    probs = probs[selection]
    bboxes = bboxes[selection]

    # Third 'output' stage. Additional refinement and landmark location.
    # ------------------------------------------------------------------
    probs, bboxes, landmarks = output_proposals(
        pyramid, bboxes, output_probability_threshold
    )

    # Remove overlapping candidates with NMS
    selection = non_maximum_suppression(probs, bboxes, output_maximum_iou)
    probs = probs[selection]
    bboxes = bboxes[selection]
    landmarks = landmarks[selection]

    return probs, bboxes, landmarks


class DetectedFaces(NamedTuple):
    """
    The output of :py:func:`detect_faces`.
    """

    probabilities: NDArray
    """
    A (num_faces, ) shaped array giving the probability score (0.0 - 1.0)
    assigned to each detected face.
    """

    angles: NDArray
    """
    A (num_faces, ) shaped array giving the angle at which the image was
    rotated (in radians) when a given face was detected.
    """

    bounding_boxes: NDArray
    """
    A (num_faces, 4) shaped array giving the coordinates of a bounding box for
    each detected face as x1, y1, x2 and y2. Coordinates are given in terms of
    input pixels and are *inclusive*.
    
    Note that the coordinates given are the un-rotated coordinates (i.e. those
    in the un-rotated input image). When angle != 0, you will need to rotate
    the image (and these coordinates) to extract a face accordingly.
    """

    landmarks: NDArray
    """
    A (num_faces, 10) shaped array giving the coordinates of five detected facial
    landmarks in each face. These are given as a series of pixel coordinates
    (x1, y1, x2, y2, ...) with the landmarks being:
    
    * Left eye
    * Right eye
    * Nose
    * Left of mouth
    * Right of mouth
    """

    images: list[Image.Image] | None
    """
    If requested, includes a cropped image of each face (rotated to the angle
    at which it was recognised by the model).
    
    All images are square-cropped by expanding the bounding box area into a
    square. To obtain just the bounding-box area, you will need to further crop
    these images according to the aspect ratio of the corresponding bounding
    box.
    """


def translate(coords: NDArray, dx: float, dy: float) -> NDArray:
    """
    Apply a translation to a set of coordinates.

    Parameters
    ==========
    coords : array (..., even-number-of-coords)
        The Input coordintaes in the final dimension (where even numbered
        entries are x-coordinates and odd numbered entries are y-coordinates.
    dx : float
    dy : float
        The translation to be applied
    """
    out = coords.copy()
    out.T[0::2] += dx
    out.T[1::2] += dy
    return out


def rotate(coords: NDArray, angle: float) -> NDArray:
    """
    Rotate a collection of coordinates around the origin, counter-clockwise.

    Parameters
    ==========
    coords : array (..., even-number-of-coords)
        The Input coordintaes in the final dimension (where even numbered
        entries are x-coordinates and odd numbered entries are y-coordinates.
    angle : float
        Angle (in radians) to rotate the coordinates by.
    """
    rotation = np.array(
        [
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle), np.cos(angle)],
        ]
    )

    # (..., 2N) -> (..., N, 2)
    coords = coords.reshape(*coords.shape[:-1], coords.shape[-1] // 2, 2)
    # (..., N, 2) -> (..., 2, N)
    coords = np.moveaxis(coords, -1, -2)

    # Perform the actual rotation
    out = rotation @ coords  # (..., 2, N)

    # (..., 2, N) -> (..., N, 2)
    out = np.moveaxis(out, -1, -2)
    # (..., N, 2) -> (..., 2N)
    out = out.reshape(*out.shape[:-2], out.shape[-2] * 2)

    return out


def detect_faces(
    image: Image.Image,
    pyramid_downscale_factor: float = np.sqrt(2),
    rotations: NDArray = np.array([-np.pi / 2, -np.pi / 4, 0.0, np.pi / 4, np.pi / 2]),
    proposal_skip_downscale_factor: float = 4.0,
    proposal_probability_threshold: float = 0.7,
    refine_probability_threshold: float = 0.7,
    output_probability_threshold: float = 0.7,
    proposal_maximum_iou: float = 0.5,
    refine_maximum_iou: float = 0.5,
    output_maximum_iou: float = 0.5,
    rotation_maximum_iou: float = 0.25,
    extract_images: int | None = 160,
) -> DetectedFaces:
    """
    Detect faces within an image using the MTCNN algorithm by Zhang et al.

    Parameters
    ==========
    image : Image
        The image to detect faces within.
    pyramid_downscale_factor : float
        The image will be processed at a range of progressively sizes, related
        by this factor.
    rotations : [radians, ...]
        The angles (in radians) at which to rotate the input image when searching.
    proposal_skip_downscale_factor : float
        Skips running the proposal filtering process on images below this
        scaling factor. The choice of '4' as a default prevents the use of
        up-scaled faces in the refinement and output stages (whose input sizes
        are twice and four times that of the proposal stage).
    proposal_probability_threshold : float
    refine_probability_threshold : float
    output_probability_threshold : float
        The is-face probability threshold for a face to be accepted during each
        of the three filtering stages.
    proposal_maximum_iou : float
    refine_maximum_iou : float
    output_maximum_iou : float
    rotation_maximum_iou : float
        The threshold of the intersection-over-union above which two rectangles
        are duplicates during each of the three processing stages and the final
        combining of rotated detections.
    extract_images : int or None
        If a number of pixels is given, includes a
        extract_images-by-extract_images pixel square-crop of each detected
        face in the output. If None, face crops are not extracted.

        The default size of 160x160 pixels is used by the FaceNet-PyTorch
        implementation of the FaceNet model. This differs from the original
        FaceNet/Inception-ResNet-v1 paper which is shown using an input size of
        299x299 pixels. We use the former size by default to better match
        FaceNet-PyTorch from which we're re-using the weights.

    Returns
    =======
    DetectedFaces
        For each detected face, a probability, angle, bounding box, set of
        facial landmarks and (optionally) a cropped image.
    """
    all_pyramids = []
    all_probs = []
    all_angles = []
    all_bboxes = []
    all_bboxes_normalised = (
        []
    )  # Internal use: Corners of a square centered on the detected face
    all_bboxes_rotated = []  # Internal use: Bounding box in rotated image coordinates
    all_landmarks = []

    # Run face detection on all rotations
    for angle_rad in rotations:
        angle_deg = (angle_rad / np.pi) * 180
        if angle_rad != 0:
            rotated_image = image.rotate(
                angle_deg, resample=Image.Resampling.BICUBIC, expand=True
            )
        else:
            rotated_image = image

        pyramid = ImagePyramid(
            rotated_image,
            min_size=12,
            downscale_factor=pyramid_downscale_factor,
        )

        probs, bboxes, landmarks = detect_faces_single_orientation(
            pyramid=pyramid,
            proposal_skip_downscale_factor=proposal_skip_downscale_factor,
            proposal_probability_threshold=proposal_probability_threshold,
            refine_probability_threshold=refine_probability_threshold,
            output_probability_threshold=output_probability_threshold,
            proposal_maximum_iou=proposal_maximum_iou,
            refine_maximum_iou=refine_maximum_iou,
            output_maximum_iou=output_maximum_iou,
        )

        # NB: bounding-box and landmark coordinates rotated into original image
        # coordinates.
        all_pyramids.append([pyramid] * len(probs))
        all_angles.append(np.full(probs.shape, angle_rad))
        all_probs.append(probs)
        all_bboxes.append(
            translate(
                rotate(
                    translate(
                        bboxes,
                        -rotated_image.size[0] / 2,
                        -rotated_image.size[1] / 2,
                    ),
                    angle_rad,
                ),
                image.size[0] / 2,
                image.size[1] / 2,
            )
        )
        all_landmarks.append(
            translate(
                rotate(
                    translate(
                        landmarks,
                        -rotated_image.size[0] / 2,
                        -rotated_image.size[1] / 2,
                    ),
                    angle_rad,
                ),
                image.size[0] / 2,
                image.size[1] / 2,
            )
        )

        # For the purposes of later NMS, we normalise the detected bounding
        # boxes to squares oriented at 0 degrees centered on their original
        # location. This isn't ideal but it is easy!
        longest_sides = bounding_boxes_to_longest_sides(bboxes)
        centers = translate(
            rotate(
                translate(
                    bounding_boxes_to_centers(bboxes),
                    -rotated_image.size[0] / 2,
                    -rotated_image.size[1] / 2,
                ),
                angle_rad,
            ),
            image.size[0] / 2,
            image.size[1] / 2,
        )
        all_bboxes_normalised.append(
            centers_to_square_bounding_boxes(centers, longest_sides)
        )

        # Bounding-box coordinates in rotated image domain (used to extract
        # face crops)
        all_bboxes_rotated.append(bboxes)

    # Combine results
    pyramids = list(chain(*all_pyramids))
    probs = np.concatenate(all_probs)
    angles = np.concatenate(all_angles)
    bboxes = np.concatenate(all_bboxes)
    bboxes_normalised = np.concatenate(all_bboxes_normalised)
    bboxes_rotated = np.concatenate(all_bboxes_rotated)
    landmarks = np.concatenate(all_landmarks)

    # Remove duplicates
    selection = non_maximum_suppression(probs, bboxes_normalised, rotation_maximum_iou)

    pyramids = [pyramids[i] for i in selection]
    probs = probs[selection]
    angles = angles[selection]
    bboxes = bboxes[selection]
    bboxes_rotated = bboxes_rotated[selection]
    landmarks = landmarks[selection]

    # Extract face images from input
    images: list[Image.Image] | None = None
    if extract_images is not None:
        images = [
            pyramid.extract((x1, y1, x2, y2), (extract_images, extract_images))
            for pyramid, (x1, y1, x2, y2) in zip(pyramids, make_square(bboxes_rotated))
        ]

    return DetectedFaces(probs, angles, bboxes, landmarks, images)



"""
Convert the FaceNet weights from FaceNet-PyTorch's format into the structure used
by this library.
"""

from typing import Any

from argparse import ArgumentParser

import torch

import numpy as np
from numpy.typing import NDArray

from itertools import count

import weightie

from faceie.nn_arguments import (
    LinearWeights,
    Conv2DWeights,
    BatchNormalisationWeights,
)

from faceie.facenet.model import (
    FaceNetWeights,
    StemWeights,
    ConvolutionalUnitWeights,
    InceptionResNetAWeights,
    ReductionAWeights,
    InceptionResNetBWeights,
    ReductionBWeights,
    InceptionResNetCWeights,
)


def n(tensor: torch.Tensor) -> NDArray:
    """
    Convert a PyTorch tensor into a numpy array.
    """
    return tensor.detach().numpy().astype(np.float32)


def extract_convolutional_unit_weights(
    unit: Any,  # facenet_pytorch.models.inception_resnet_v1.BasicConv2d
) -> ConvolutionalUnitWeights:
    return ConvolutionalUnitWeights(
        kernel=n(unit.conv.weight),
        batch_normalisation=BatchNormalisationWeights(
            population_mean=n(unit.bn.running_mean),
            population_variance=n(unit.bn.running_var),
            weights=n(unit.bn.weight),
            biases=n(unit.bn.bias),
            # Hard-coded in FaceNet-PyTorch to match TensorFlow but not stored
            # in the serialised state_dict.
            eps=0.001,
        ),
    )


def extract_convolution_weights(unit: torch.nn.Conv2d) -> Conv2DWeights:
    assert unit.bias is not None
    return Conv2DWeights(
        weights=n(unit.weight),
        biases=n(unit.bias),
    )


def extract_weights(
    model: Any,  # facenet_pytorch.models.inception_resnet_v1.InceptionResnetV1
) -> FaceNetWeights:
    return FaceNetWeights(
        stem=StemWeights(
            extract_convolutional_unit_weights(model.conv2d_1a),
            extract_convolutional_unit_weights(model.conv2d_2a),
            extract_convolutional_unit_weights(model.conv2d_2b),
            extract_convolutional_unit_weights(model.conv2d_3b),
            extract_convolutional_unit_weights(model.conv2d_4a),
            extract_convolutional_unit_weights(model.conv2d_4b),
        ),
        inception_resnet_a=[
            InceptionResNetAWeights(
                branch_1_conv=extract_convolutional_unit_weights(module.branch0),
                branch_2_conv_1=extract_convolutional_unit_weights(module.branch1[0]),
                branch_2_conv_2=extract_convolutional_unit_weights(module.branch1[1]),
                branch_3_conv_1=extract_convolutional_unit_weights(module.branch2[0]),
                branch_3_conv_2=extract_convolutional_unit_weights(module.branch2[1]),
                branch_3_conv_3=extract_convolutional_unit_weights(module.branch2[2]),
                output_conv=extract_convolution_weights(module.conv2d),
                # Not stored in the serialised state_dict (but defined in
                # FaceNet-PyTorch)
                residual_scale=0.17,
            )
            for module in model.repeat_1
        ],
        reduction_a=ReductionAWeights(
            branch_1_conv=extract_convolutional_unit_weights(model.mixed_6a.branch0),
            branch_2_conv_1=extract_convolutional_unit_weights(
                model.mixed_6a.branch1[0]
            ),
            branch_2_conv_2=extract_convolutional_unit_weights(
                model.mixed_6a.branch1[1]
            ),
            branch_2_conv_3=extract_convolutional_unit_weights(
                model.mixed_6a.branch1[2]
            ),
        ),
        inception_resnet_b=[
            InceptionResNetBWeights(
                branch_1_conv=extract_convolutional_unit_weights(module.branch0),
                branch_2_conv_1=extract_convolutional_unit_weights(module.branch1[0]),
                branch_2_conv_2=extract_convolutional_unit_weights(module.branch1[1]),
                branch_2_conv_3=extract_convolutional_unit_weights(module.branch1[2]),
                output_conv=extract_convolution_weights(module.conv2d),
                # Not stored in the serialised state_dict (but defined in
                # FaceNet-PyTorch)
                residual_scale=0.10,
            )
            for module in model.repeat_2
        ],
        reduction_b=ReductionBWeights(
            branch_1_conv_1=extract_convolutional_unit_weights(
                model.mixed_7a.branch0[0]
            ),
            branch_1_conv_2=extract_convolutional_unit_weights(
                model.mixed_7a.branch0[1]
            ),
            branch_2_conv_1=extract_convolutional_unit_weights(
                model.mixed_7a.branch1[0]
            ),
            branch_2_conv_2=extract_convolutional_unit_weights(
                model.mixed_7a.branch1[1]
            ),
            branch_3_conv_1=extract_convolutional_unit_weights(
                model.mixed_7a.branch2[0]
            ),
            branch_3_conv_2=extract_convolutional_unit_weights(
                model.mixed_7a.branch2[1]
            ),
            branch_3_conv_3=extract_convolutional_unit_weights(
                model.mixed_7a.branch2[2]
            ),
        ),
        inception_resnet_c=(
            [
                InceptionResNetCWeights(
                    branch_1_conv=extract_convolutional_unit_weights(module.branch0),
                    branch_2_conv_1=extract_convolutional_unit_weights(
                        module.branch1[0]
                    ),
                    branch_2_conv_2=extract_convolutional_unit_weights(
                        module.branch1[1]
                    ),
                    branch_2_conv_3=extract_convolutional_unit_weights(
                        module.branch1[2]
                    ),
                    output_conv=extract_convolution_weights(module.conv2d),
                    # Not stored in the serialised state_dict (but defined in
                    # FaceNet-PyTorch)
                    residual_scale=0.20,
                )
                for module in model.repeat_3
            ]
            + [
                InceptionResNetCWeights(
                    branch_1_conv=extract_convolutional_unit_weights(
                        model.block8.branch0
                    ),
                    branch_2_conv_1=extract_convolutional_unit_weights(
                        model.block8.branch1[0]
                    ),
                    branch_2_conv_2=extract_convolutional_unit_weights(
                        model.block8.branch1[1]
                    ),
                    branch_2_conv_3=extract_convolutional_unit_weights(
                        model.block8.branch1[2]
                    ),
                    output_conv=extract_convolution_weights(model.block8.conv2d),
                    # Not stored in the serialised state_dict (but defined in
                    # FaceNet-PyTorch)
                    residual_scale=1.0,
                )
            ]
        ),
        output_dimension_reduction=LinearWeights(
            # NB: Weights stored transposed in PyTorch
            weights=n(model.last_linear.weight.T),
            biases=None,
        ),
        output_batch_normalisation=BatchNormalisationWeights(
            population_mean=n(model.last_bn.running_mean),
            population_variance=n(model.last_bn.running_var),
            weights=n(model.last_bn.weight),
            biases=n(model.last_bn.bias),
            # Hard-coded in FaceNet-PyTorch to match TensorFlow but not stored
            # in the serialised state_dict.
            eps=0.001,
        ),
    )


class StateDictAsObjects:
    """
    Present a PyTorch state dictionary (whose keys look like, e.g.
    repeat_3.4.branch1.2.bn.weight) as if it was an instantiated PyTorch module
    (e.g. make it accessible like this.repeat_3[4]branch1[2]bn.weight)
    """

    def __init__(self, state_dict: dict[str, Any], prefix: str = "") -> None:
        self._state_dict = state_dict
        self._prefix = prefix

    def __getattr__(self, key) -> Any:
        full_key = self._prefix + key

        # Try finding a leaf value with the correct name
        try:
            return self._state_dict[full_key]
        except KeyError:
            pass

        # Failing that, try to produce a nesting
        new_prefix = full_key + "."
        if any(k.startswith(new_prefix) for k in self._state_dict):
            return StateDictAsObjects(self._state_dict, new_prefix)

        raise AttributeError(key)

    def __getitem__(self, index) -> Any:
        try:
            return self.__getattr__(str(index))
        except AttributeError:
            raise IndexError(index)

    def __iter__(self) -> Any:
        for i in count():
            try:
                yield self[i]
            except IndexError:
                break


def main():
    parser = ArgumentParser(
        description="""
        Convert FaceNet weights from FaceNet-PyTorch format to that used by
        faceie.
        """
    )
    parser.add_argument("input", help="FaceNet-PyTorch weights file")
    parser.add_argument("output", help="Output (faceie-format weights) filename.")

    args = parser.parse_args()

    state_dict = torch.load(args.input)
    weights = extract_weights(StateDictAsObjects(state_dict))

    with open(args.output, "wb") as f:
        weightie.dump(weights, f)



from PIL import Image

import numpy as np
from numpy.typing import NDArray


def image_to_array(image: Image.Image) -> NDArray:
    """
    Convert an 8-bit RGB PIL image into a float32 Numpy array with shape (3,
    height, width) and valuesin the range -1 to +1.
    """
    out = np.asarray(image).astype(np.float32)

    # Rescale
    out -= 127.5
    out /= 128.0

    # Move channels to front
    out = np.moveaxis(out, 2, 0)

    return out



"""
A collection of generic deep neural network functions.
"""

from enum import Enum, auto

import numpy as np
from numpy.typing import NDArray

from math import ceil, floor


def linear(x: NDArray, weights: NDArray, biases: NDArray | None) -> NDArray:
    """
    A simple linear/dense layer: (x @ weights) + biases
    """
    x = x @ weights

    if biases is not None:
        x += biases

    return x


def l2_normalisation(x: NDArray, axis: int, eps: float = 1e-12) -> NDArray:
    """
    Scale the values along the given axis of X such that their squares sum to
    one.

    Parameters
    ==========
    x : array (...)
    axis : int
        The axis in 'x' to normalise.
    eps : float
        To avoid division by zero in the case where the squared values sum to
        zero (or almost zero), a minimum value to scale by instead.

        Default value matches the default of PyTorch's nn.functional.normalize

    Returns
    =======
    array (...)
        Same shape as 'x'.
    """
    norm = np.linalg.norm(x, axis=axis, keepdims=True)
    norm = np.maximum(norm, eps)
    return x / norm


def batch_normalisation_2d(
    x: NDArray,
    population_mean: NDArray,
    population_variance: NDArray,
    weights: NDArray,
    biases: NDArray,
    eps: float = 1e-5,
    channel_axis: int = 1,
) -> NDArray:
    """
    Apply a (precomputed) batch normalisation to a collection of 2D
    multi-channel values.

    See "Batch Normalization: Accelerating Deep Network Training by Reducing
    Internal Covariate Shift" by Ioffe and Szegedy (2015).

    Parameters
    ==========
    x : array (..., channel, ...)
        The input array, typically of shape (batch, channel, height, width).
    population_mean : array (channel, )
    population_variance : array (channel, )
        The (approximated) population mean and variance measured during
        training. Given for each channel as a whole in the input.
    weights : array (channel, )
    biases : array (channel, )
        A 1D array giving the scaling factors and offsets to apply to all
        values in each channel in the input.
    eps : float
        The standard deviation is actually computed as sqrt(variance + eps)
        where eps is a small constant used to avoid division by zero.
    channel_axis : int
        The index of the channel axis in ``x``.
    """
    # For the sake of avoiding scaling and offsetting the whole input twice
    # (once to normalise and again to apply the weights and biases), we can
    # fold these together into a single scaling and offsetting operaiton.
    #
    # Rarranging from the naive operation:
    #
    #   std = sqrt(variance + eps)
    #
    #   out = ( ((x - mean) / std)                     * weights) + biases
    #          (((x - mean) / std) + (biases/weights)) * weights
    #
    #          (((x - mean                         ) / std) + (biases/weights)) * weights
    #           ((x - mean + ((biases/weights)*std)) / std)                     * weights
    #
    #          ((x - mean + ((biases/weights)*std)) /  std           ) * weights
    #           (x - mean + ((biases/weights)*std)) / (std / weights)
    #
    # Given this factorisation we can rewrite the whole thing as
    #
    #   out = (x + offset) * scale
    #
    # Where
    #
    #   scale = weights / std
    #
    #   offset = ((biases/weights)*std) - mean
    #   offset = (biases*(std/weights)) - mean
    #   offset = (biases/scale        ) - mean
    #
    # TODO: Perhaps we ought to consider performing this folding in a helper
    # function and just accepting the scale and offset values as arguments...?

    population_std = np.sqrt(population_variance + eps)
    scale = weights / population_std
    offset = (biases / scale) - population_mean

    # Reshape scale and offset to broadcast over the channels.
    channel_axis %= x.ndim  # Make positive
    new_shape = tuple(-1 if i == channel_axis else 1 for i in range(x.ndim))
    scale = scale.reshape(new_shape)
    offset = offset.reshape(new_shape)

    # Scale and offset according to learned parameters
    x = (x + offset) * scale

    return x


class PaddingType(Enum):
    """
    'Automatic' padding types for convolution-style functions.
    """

    valid = auto()
    """
    Alias for padding = 0, that is, only use 'valid' input data.
    """

    same = auto()
    """
    Use the ammount of padding required to make the convolution output the same
    size as the input. Requires that the kernel has odd dimensions.
    """


def conv2d(
    img: NDArray,
    weights: NDArray,
    biases: NDArray | None = None,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] | PaddingType = 0,
) -> NDArray:
    """
    Perform 2D convolution on the provided image using the provided kernel
    weights and biases.

    Parameters
    ==========
    img : array (in_channels, img_height, img_width) or (num_batches, in_channels, img_height, img_width)
        The input image, or a batch of images to process simultaneously.
    weights : array (out_channels, in_channels, kernel_height, kernel_width)
        The kernel to apply to the input.
    biases : array (out_channels, ) or None
        The biases to add to each output channel -- or None to not add biases.
    stride : int or (int, int)
        The stride of the convolutional filter.
    padding : int or (int, int) or PaddingType
        The (zero) padding to add to the top-and-bottom and left-and-right of
        the input. When non-zero, stride must be 1.

    Returns
    =======
    array (out_channels, out_height, out_width) or (num_batches, out_channels, out_height, out_width)
        The convolution output.

        The num_batches dimension will be present iff it it was included in the
        input img.

        The output dimensions are given by the formulae below:

        * out_height = floor((img_height + 2*padding[0] - kernel_size[0]) / stride[0]) + 1
        * out_width = floor((img_width + 2*padding[1] - kernel_size[1]) / stride[1]) + 1
    """
    out_channels, in_channels, kernel_height, kernel_width = weights.shape

    # Expand stride to pairs
    if isinstance(stride, int):
        stride = (stride, stride)

    # Expand padding to pairs
    if isinstance(padding, PaddingType):
        if padding is PaddingType.valid:
            padding = 0
        elif padding is PaddingType.same:
            padding = (
                (kernel_height - 1) // 2,
                (kernel_width - 1) // 2,
            )
    if isinstance(padding, int):
        padding = (padding, padding)

    # Sanity check stride/padding not used simultaneously
    if (stride[0] != 1 or stride[1] != 1) and (padding[0] != 0 or padding[1] != 0):
        raise ValueError("Padding not supported when stride is not 1.")

    # Internally always treat the input as batched
    batched = len(img.shape) == 4
    if not batched:
        img = np.expand_dims(img, 0)

    # Add padding (if required).
    #
    # NB: Experimentally, performing this copy is no slower than implementing
    # the edge-cases manually below but makes for a substantially simpler
    # implementation. (See commit 6892113 for a special-case based
    # implementation of this function.
    if padding[0] or padding[1]:
        img = np.pad(
            img,
            (
                (0, 0),  # num_batches
                (0, 0),  # in_channels
                (padding[0], padding[0]),  # height
                (padding[1], padding[1]),  # width
            ),
        )

    # Produce a sliding window over the input image to which the kernel will be
    # applied.
    #
    # NB: It is unclear why mypy fails to find a suitable override of
    # sliding_window_view matching our (pretty vanilla) usage here. As such we
    # just ignore it for now...
    windows = np.lib.stride_tricks.sliding_window_view(  # type: ignore
        img,  # (num_batches, in_channels, padded_img_height, padded_img_width)
        (kernel_height, kernel_width),
        axis=(-2, -1),
    )  # (num_batches, in_channels, out_height, out_width, kernel_height, kernel_width)

    # Apply stride
    windows = windows[:, :, :: stride[0], :: stride[1], :, :]

    num_batches = windows.shape[0]
    out_height, out_width = windows.shape[2:4]

    # Perform convolution
    out = np.tensordot(
        weights,  # (out_channels, in_channels, kernel_height, kernel_width)
        windows,  # (num_batches, in_channels, out_height, out_width, kernel_height, kernel_width)
        axes=((1, 2, 3), (1, 4, 5)),
    )  # (out_channels, num_batches, out_height, out_width)
    out = np.moveaxis(out, 0, 1)  # (num_batches, out_channels, out_height, out_width)

    # Add biases (if given)
    if biases is not None:
        out += np.expand_dims(biases, (1, 2))  # bases.shape = (out_channels, 1, 1)

    if batched:
        return out
    else:
        return out[0]


def relu(x: NDArray) -> NDArray:
    """
    Implements a Rectified Linear Unit (ReLU).

    Values in ``x`` are mapped to themselves if greater than zero or clamped to
    zero otherwise.
    """
    return np.maximum(x, 0)


def prelu(x: NDArray, parameters: NDArray, axis: int) -> NDArray:
    """
    Implements Parameterised Rectified Linear Unit (PReLU).

    Values in ``x`` are mapped to themselves if greater than zero or scaled by
    the value in ``parameters`` if less than zero.

    Parameters
    ==========
    x : array (any shape)
        The input array.
    parameters : array (1-dimensional array)
        The scaling factors used for negative values in x. The scaling value in
        ``parameters`` used for a given value in ``x`` is based on its location
        along the axis identified by the ``axis`` parameter.

        This array must have the same length as the chosen axis in ``x``.
    axis : int
        The axis index in ``x`` to use to select the scaling parameter in
        ``parameters``.

    Returns
    =======
    array (same shape as ``x``)
    """
    # Reshape to broadcast to the selected axis
    parameters = parameters.reshape(*(-1 if i == axis else 1 for i in range(x.ndim)))

    # NB: This method is kind of slow compared to PyTorch's implementation
    # which (I imagine) has a nicely vectorised compare-and-scale routine going
    # on.
    #
    # Depending on how your measure it, this code runs between 2 and 10 times
    # slower(!)
    return x + ((x < 0) * x * (parameters - 1))


def max_pool_2d(
    x: NDArray,
    kernel: int | tuple[int, int],
    stride: int | tuple[int, int],
    ceil_mode: bool = False,
    out: NDArray | None = None,
) -> NDArray:
    """
    Apply 2D 'max pooling' convolution filter in which the input is convolved
    using the 'max' function acting over kernel-sized regions.

    .. warning::

        This function is significantly slower than similar functions in (e.g.)
        PyTorch due to numpy's apparent lack of SIMD or multicore support in
        its 'amax' function.

    Parameters
    ==========
    x : array (..., in_height, in_width)
        The input array. The final two dimensions are used as 2D coordinates.
        They must be exact multiples of block_height and block_width.
    kernel : int or (int, int)
        The kernel height and width.
    stride : int or (int, int)
        The step size.
    ceil_mode : bool
        If True, add -inf padding to the right and bottom edges of the input
        when necessary to ensure that all input values are represented in the
        output.

        For some combinations of input size, kernel and stride, it is possible
        for some input values to be omitted from processing. The illustration
        below shows the regions of an input processed by a kernel::

            +---+---+---+-+   ceil_mode == False
            |   |   |   |X|
            +---+---+---+X|   X = values not processed by any kernel
            |   |   |   |X|
            +---+---+---+X|
            +XXXXXXXXXXXXX+

        Here, because the kernel and stride do not exactly divide the input,
        some inputs at the bottom and right sides are not processed by any
        kernels and are effectively ignored. This is the behaviour when
        ``ceil_mode`` is False.

        When ``ceil_mode`` is True, the input is effectively padded with -inf
        to enable an extra kernel to fit, capturing the edge-most values::

            +---+---+---+---+   ceil_mode == True
            |   |   |   | PP|
            +---+---+---+---+   P = padding '-inf' values added to input
            |   |   |   | PP|
            +---+---+---+---+
            |   |   |   | PP|
            +PPP+PPP+PPP+PPP+

        The result is that the output size grows by one and all input values
        are processed by at least one kernel. (Unless of course the stride is
        larger than the kernel!)

        .. note::

            The case where the stride is larger than the kernel is considered
            degenerate in that inputs will be ignored between kernels which fit
            in the input without padding. In this case, ceil_mode will continue
            to expand the output

    out: NDArray or None
        If given, an array into which to write the output (see return value for
        expected size). Otherwise, a new array will be allocated.

    Returns
    =======
    array (..., out_height, out_width)
        The output of the convolution.

        The output dimensions are:

            out_height = ceil(((height - kernel[0])/stride[0]) + 1)
            out_width = ceil(((width - kernel[1])/stride[1]) + 1)

        (Replace 'ceil' with 'floor' if ceil_mode is False.)
    """
    # Expand kernel/stride size shorthands
    if isinstance(kernel, int):
        kernel = (kernel, kernel)
    if isinstance(stride, int):
        stride = (stride, stride)

    # If the kernel is larger than the input then there are enough tricks
    # needed below to make it not worth handling this edge-case until I
    # actually need to...
    if kernel[0] > x.shape[-2]:
        raise ValueError("Kernel taller than input.")
    if kernel[1] > x.shape[-1]:
        raise ValueError("Kernel wider than input.")

    # We don't have any sensible handling for when ceil_mode is used with a
    # kernel smaller than the step -- what happens in this case is somewhat
    # ill-defined anyway...
    if ceil_mode:
        if stride[0] > kernel[0]:
            raise ValueError("Stride taller than kernel.")
        if stride[1] > kernel[1]:
            raise ValueError("Stride wider than kernel.")

    # Get a windowed view of 'x's final two dimensions.
    #
    # NB: Regardless of the ceil_mode setting, this includes only 'complete'
    # kernel inputs where no padding is required. Padded cases are handled
    # specially later.
    #
    # windowed.shape == (..., wh, ww, kernel[0], kernel[1])
    windowed = np.lib.stride_tricks.sliding_window_view(
        x,
        window_shape=kernel,
        axis=(-2, -1),
    )  # type: ignore
    # XXX: Unclear why MyPy doesn't like the type of 'x' above here...

    # Pull out required strides (again ignoring 'incomplete' windows)
    #
    # (..., out_height_complete, out_width_complete, kernel[0], kernel[1])
    windowed = windowed[..., :: stride[0], :: stride[1], :, :]

    # Create the output array (including an extra row or column as needed
    # when ceil_mode is True).
    #
    # (..., out_height, out_width)
    rounding_mode = ceil if ceil_mode else floor
    out_height = rounding_mode(((x.shape[-2] - kernel[0]) / stride[0]) + 1)
    out_width = rounding_mode(((x.shape[-1] - kernel[1]) / stride[1]) + 1)
    if out is None:
        out = np.zeros(shape=(x.shape[:-2] + (out_height, out_width)), dtype=x.dtype)
    else:
        # Sanity check user provided correctly shaped output array
        assert out.shape == x.shape[:-2] + (out_height, out_width)

    # We now perform max pooling for all complete kernels
    np.amax(
        windowed,
        axis=(-2, -1),
        out=out[..., : windowed.shape[-4], : windowed.shape[-3]],
    )  # (..., out_height_complete, out_width_complete)

    # Now we perform max pooling for any overspilled kernels when required in
    # ceil_mode
    if ceil_mode:
        # When some partial kernels are required, find their size
        bottommost_kernel_end = ((out.shape[-2] - 1) * stride[0]) + kernel[0]
        rightmost_kernel_end = ((out.shape[-1] - 1) * stride[1]) + kernel[1]

        stragglers_bottom = stragglers_right = 0
        if bottommost_kernel_end > x.shape[-2]:
            stragglers_bottom = x.shape[-2] - (bottommost_kernel_end - kernel[0])
        if rightmost_kernel_end > x.shape[-1]:
            stragglers_right = x.shape[-1] - (rightmost_kernel_end - kernel[1])

        # Process straggler rows and columns which were not processed by any
        # full-sized kernel.
        if stragglers_bottom:
            max_pool_2d(
                x[..., -stragglers_bottom:, :],
                kernel=(stragglers_bottom, kernel[1]),
                stride=stride,
                ceil_mode=False,
                out=out[..., -1:, : windowed.shape[-3]],
            )

        if stragglers_right:
            max_pool_2d(
                x[..., :, -stragglers_right:],
                kernel=(kernel[0], stragglers_right),
                stride=stride,
                ceil_mode=False,
                out=out[..., : windowed.shape[-4], -1:],
            )

        if stragglers_bottom and stragglers_right:
            np.max(
                x[..., -stragglers_bottom:, -stragglers_right:],
                axis=(-2, -1),
                out=out[..., -1, -1],
            )
    else:
        # Sanity check when ceil_mode is False
        assert out.shape == windowed.shape[:-2]

    return out


def softmax(x: NDArray, axis: int = -1) -> NDArray:
    """
    A numerically stable implementation of the soft(arg)max function.
    """
    # For reasons of numerical stability, offset all values such that we don't
    # inadvertently produce an overflow after exponentiation.
    #
    # NB: The subtraction has no effect on the outcome otherwise, e.g. because:
    #
    #     e^a / (e^a + e^b) = e^(a-x)      / (e^(a-x)      + e^(b-x))
    #                       = (e^a * e^-x) / ((e^a * e^-x) + (e^b * e^-x))
    #                       = (e^a * e^-x) / ((e^a         + e^b          ) * e^-x)
    #                       =  e^a         / ( e^a         + e^b)
    #
    # NB: We perform the subtraction locally within each axis to avoid excessively scaling down
    # unrelated values which also introduces numerical stability issues but in the opposite
    # direction.
    x = x - np.max(x, axis=axis, keepdims=True)

    x = np.exp(x)
    return x / np.sum(x, axis=axis, keepdims=True)



import pytest

import numpy as np
from numpy.typing import NDArray

import torch

from faceie.nn import (
    l2_normalisation,
    PaddingType,
    batch_normalisation_2d,
    conv2d,
    relu,
    prelu,
    max_pool_2d,
    softmax,
)


def test_l2_normalisation() -> None:
    rand = np.random.RandomState(0)

    x = rand.uniform(-10, 10, size=(3, 4, 5)).astype(np.float32)

    out = l2_normalisation(x, axis=1)

    # Work out model answer using PyTorch implementation
    exp_tensor = torch.nn.functional.normalize(torch.tensor(x), p=2.0, dim=1)
    exp = exp_tensor.numpy()

    print(out)
    print(exp)

    print(np.sum(out**2, axis=1))
    print(np.sum(exp**2, axis=1))

    assert np.allclose(out, exp, atol=1e-6)


@pytest.mark.parametrize("channel_axis", [1, -3])
def test_batch_normalisation_2d(channel_axis: int) -> None:
    rand = np.random.RandomState(0)

    num_batches = 2
    num_channels = 3

    x = rand.uniform(-10, 10, size=(num_batches, num_channels, 2, 2)).astype(np.float32)

    population_mean = rand.normal(size=num_channels).astype(np.float32)
    population_variance = np.abs(rand.normal(size=num_channels).astype(np.float32))
    weights = rand.normal(size=num_channels).astype(np.float32)
    biases = rand.normal(size=num_channels).astype(np.float32)
    eps = 1e-5

    out = batch_normalisation_2d(
        x,
        population_mean,
        population_variance,
        weights,
        biases,
        eps=eps,
        channel_axis=channel_axis,
    )

    # Work out model answer using PyTorch implementation
    torch_batch_normalisation_2d = torch.nn.BatchNorm2d(num_channels, eps=eps).eval()
    with torch.no_grad():
        assert torch_batch_normalisation_2d.running_mean is not None
        assert torch_batch_normalisation_2d.running_var is not None
        torch_batch_normalisation_2d.running_mean[:] = torch.tensor(population_mean)
        torch_batch_normalisation_2d.running_var[:] = torch.tensor(population_variance)
        torch_batch_normalisation_2d.weight[:] = torch.tensor(weights)
        torch_batch_normalisation_2d.bias[:] = torch.tensor(biases)

        exp = torch_batch_normalisation_2d(torch.tensor(x)).numpy()

    print(out)
    print(exp)

    assert np.allclose(out, exp, atol=1e-6)


class TestConv2D:
    @pytest.mark.parametrize(
        "kernel, stride, padding",
        [
            # No stride or padding
            ((3, 3), 1, 0),
            # Not square
            ((3, 5), 1, 0),
            # Strided
            ((3, 3), 2, 0),
            ((3, 3), (2, 1), 0),
            ((3, 3), (1, 2), 0),
            ((3, 5), (1, 2), 0),  # Also not square!
            # Padded
            ((3, 3), 1, 1),
            ((5, 5), 1, 2),
            ((5, 5), 1, (2, 0)),
            ((5, 5), 1, (0, 2)),
            ((3, 5), 1, (0, 2)),  # Also not square!
        ],
    )
    def test_one_hot(
        self,
        kernel: tuple[int, int],
        stride: int | tuple[int, int],
        padding: int | tuple[int, int],
    ) -> None:
        # This test convolves the input with various kernel sizes, strides and
        # paddings when processing a series of one-hot input arrays. These
        # provide an easy-to-debug test case for working out issues with
        # padding and suchlike.

        weights = np.arange(1, (kernel[0] * kernel[1]) + 1, dtype=np.float32).reshape(
            1, 1, *kernel
        )
        biases = np.array([100], dtype=np.float32)

        height = 6
        width = 10

        # Compare output with PyTorch implementation
        torch_conv2d = torch.nn.Conv2d(
            in_channels=1,
            out_channels=1,
            kernel_size=kernel,
            stride=stride,
            padding=padding,
        )
        with torch.no_grad():
            torch_conv2d.weight[:] = torch.tensor(weights)
            assert torch_conv2d.bias is not None
            torch_conv2d.bias[:] = torch.tensor(biases)

        for i in range(width * height):
            img = np.zeros((1, 1, height, width), dtype=np.float32)
            img.flat[i] = 1

            print("")
            print(f"{height=}, {width=}, {kernel=}, {stride=}, {padding=}")
            print(f"{img.shape=}")
            print(img)

            out = conv2d(img, weights, biases, stride=stride, padding=padding)

            print(f"{out.shape=}")
            print(out)

            with torch.no_grad():
                exp_out = torch_conv2d(torch.tensor(img)).numpy()

            print(f"{exp_out.shape=}")
            print(exp_out)

            assert np.array_equal(out, exp_out)

    def test_batches_and_channels(self) -> None:
        # Compare behaviour against PyTorch with a testcase with random weights
        # and inputs which uses batching and  multiple input and output
        # channels.
        num_batches = 2
        in_channels = 3
        out_channels = 4
        kernel_height = 5
        kernel_width = 7

        img_height = 768
        img_width = 1024

        torch_conv2d = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=(kernel_height, kernel_width),
        )

        # Sanity check
        assert torch_conv2d.weight is not None
        assert torch_conv2d.bias is not None
        assert torch_conv2d.weight.shape == (
            out_channels,
            in_channels,
            kernel_height,
            kernel_width,
        )
        assert torch_conv2d.bias.shape == (out_channels,)

        # Get model answer
        img_tensor = torch.randn(num_batches, in_channels, img_height, img_width)
        with torch.no_grad():
            torch_out = torch_conv2d(img_tensor)

        # Convert image/weights/biases to NDArray
        img = img_tensor.numpy()
        weights = torch_conv2d.weight.detach().numpy()
        biases = torch_conv2d.bias.detach().numpy()

        out = conv2d(img, weights, biases)

        # Sanity check shape
        out_height = img_height - ((kernel_height // 2) * 2)
        out_width = img_width - ((kernel_width // 2) * 2)
        assert out.shape == (num_batches, out_channels, out_height, out_width)

        # Check equivalent to PyTorch
        #
        # NB higher tollerance due to float32 precision
        assert np.allclose(out, torch_out.numpy(), atol=1e-6)

    def test_padding_same(self) -> None:
        kernel = (5, 3)

        weights = np.zeros((1, 1, kernel[0], kernel[1]), dtype=np.float32)

        height = 6
        width = 10

        img = np.zeros((1, height, width), dtype=np.float32)

        out = conv2d(img, weights, padding=PaddingType.same)

        assert out.shape == img.shape


def test_relu() -> None:
    num_channels = 10

    input_shape = (3, num_channels, 100, 200)

    torch_relu = torch.nn.ReLU(inplace=False)

    # Get model answer
    in_tensor = torch.randn(*input_shape)
    with torch.no_grad():
        out_tensor = torch_relu(in_tensor)

    # Convert types
    input = in_tensor.numpy()

    out = relu(input)

    assert np.allclose(out, out_tensor.numpy(), atol=1e-6)


def test_prelu() -> None:
    num_channels = 10

    # Arbitrary choice except for the axis of choice being at index 1 which is
    # assumed by PyTorch.
    input_shape = (3, num_channels, 100, 200)

    torch_prelu = torch.nn.PReLU(num_channels)
    with torch.no_grad():
        # Randomise weights
        torch_prelu.weight[:] = torch.rand(*torch_prelu.weight.shape) * 2

    # Get model answer
    in_tensor = torch.randn(*input_shape)
    with torch.no_grad():
        out_tensor = torch_prelu(in_tensor)

    # Convert types
    input = in_tensor.numpy()
    parameters = torch_prelu.weight.detach().numpy()

    out = prelu(input, parameters, axis=1)

    assert np.allclose(out, out_tensor.numpy(), atol=1e-6)


@pytest.mark.parametrize("ceil_mode", [False, True])
@pytest.mark.parametrize("square", [False, True])
def test_max_pool_2d(ceil_mode: bool, square: bool) -> None:
    for stride in range(1, 5):
        for kernel in range(stride, 5):

            torch_max_pool_2d = torch.nn.MaxPool2d(
                kernel_size=kernel if square else (kernel, kernel + 1),
                stride=stride if square else (stride, stride + 1),
                ceil_mode=ceil_mode,
            )

            # Work through all a one-hot 2D arrays of the following size
            w = 8
            h = 6
            for i in range(w * h):
                ar = np.zeros(w * h, dtype=float).reshape(1, 1, h, w)
                ar.flat[i] = 1.0

                out = max_pool_2d(
                    ar,
                    kernel=kernel if square else (kernel, kernel + 1),
                    stride=stride if square else (stride, stride + 1),
                    ceil_mode=ceil_mode,
                )
                with torch.no_grad():
                    exp_out = torch_max_pool_2d(torch.tensor(ar)).numpy()

                assert np.array_equal(out, exp_out), f"{out=}\n{exp_out=}"


@pytest.mark.parametrize("dim", [0, 1, -1, -2])
def test_softmax(dim: int) -> None:
    np.random.seed(0)
    x = np.random.uniform(-10, 10, size=(3, 3))

    torch_softmax = torch.nn.Softmax(dim)
    exp = torch_softmax(torch.tensor(x)).detach().numpy()

    actual = softmax(x, axis=dim)

    print(exp)
    print(actual)

    assert np.allclose(actual, exp)



import subprocess

from pathlib import Path

import faceie


def test_mypy():
    subprocess.run(
        [
            "mypy",
            str(Path(faceie.__file__).parent),
            str(Path(__file__).parent),
        ],
        check=True,
    )
